"""Unit tests for :class:`esmf_regrid.experimental.unstructured_regrid.MeshInfo`."""

import numpy as np
from numpy import ma

from esmf_regrid import Constants
from esmf_regrid.esmf_regridder import GridInfo, Regridder
from esmf_regrid.experimental.unstructured_regrid import MeshInfo
from esmf_regrid.tests import get_result_path, make_grid_args


def _make_small_mesh_args():
    ugrid_node_coords = np.array(
        [[0.0, 0.0], [0.0, 10.0], [10.0, 0.0], [10.0, 10.0], [10.0, 20.0]]
    )
    ugrid_face_node_connectivity = ma.array(
        [[0, 2, 3, -1], [3, 0, 1, 4]],
        mask=np.array([[0, 0, 0, 1], [0, 0, 0, 0]]),
    )
    node_start_index = 0
    return ugrid_node_coords, ugrid_face_node_connectivity, node_start_index


def test_make_mesh():
    """Basic test for creating :meth:`~esmf_regrid.esmf_regridder.GridInfo.make_esmf_field`."""
    coords, nodes, _ = _make_small_mesh_args()
    mesh_0 = MeshInfo(coords, nodes, 0)
    esmf_mesh_0 = mesh_0.make_esmf_field()
    esmf_mesh_0.data[:] = 0

    relative_path = (
        "experimental",
        "unstructured_regrid",
        "test_MeshInfo",
        "small_mesh.txt",
    )
    fname = get_result_path(relative_path)
    with open(fname) as file:
        expected_repr = file.read()

    one_indexed_nodes = nodes + 1
    mesh_1 = MeshInfo(coords, one_indexed_nodes, 1)
    esmf_mesh_1 = mesh_1.make_esmf_field()
    esmf_mesh_1.data[:] = 0

    assert esmf_mesh_0.__repr__() == esmf_mesh_1.__repr__() == expected_repr


def test_regrid_with_mesh():
    """Basic test for regridding with :meth:`~esmf_regrid.esmf_regridder.GridInfo.make_esmf_field`."""
    mesh_args = _make_small_mesh_args()
    mesh = MeshInfo(*mesh_args)

    grid_args = make_grid_args(2, 3)
    grid = GridInfo(*grid_args)

    mesh_to_grid_regridder = Regridder(mesh, grid)
    mesh_input = np.array([3, 2])
    grid_output = mesh_to_grid_regridder.regrid(mesh_input)
    expected_grid_output = np.array(
        [
            [2.671294712940605, 3.0],
            [2.0885553467353097, 2.9222786250561574],
            [2.0, 2.3397940801753307],
        ]
    )
    assert ma.allclose(expected_grid_output, grid_output)

    grid_to_mesh_regridder = Regridder(grid, mesh)
    grid_input = np.array([[0, 0], [1, 0], [2, 1]])
    mesh_output = grid_to_mesh_regridder.regrid(grid_input)
    expected_mesh_output = np.array([0.1408245341331448, 1.19732762534643])
    assert ma.allclose(expected_mesh_output, mesh_output)

    def _give_extra_dims(array):
        result = np.stack([array, array + 1])
        result = np.stack([result, result + 10, result + 100])
        return result

    extra_dim_mesh_input = _give_extra_dims(mesh_input)
    extra_dim_grid_output = mesh_to_grid_regridder.regrid(extra_dim_mesh_input)
    extra_dim_expected_grid_output = _give_extra_dims(expected_grid_output)
    assert ma.allclose(extra_dim_expected_grid_output, extra_dim_grid_output)

    extra_dim_grid_input = _give_extra_dims(grid_input)
    extra_dim_mesh_output = grid_to_mesh_regridder.regrid(extra_dim_grid_input)
    extra_dim_expected_mesh_output = _give_extra_dims(expected_mesh_output)
    assert ma.allclose(extra_dim_expected_mesh_output, extra_dim_mesh_output)


def test_regrid_bilinear_with_mesh():
    """Basic test for regridding with :meth:`~esmf_regrid.esmf_regridder.GridInfo.make_esmf_field`."""
    # We create a mesh with the following shape:
    # 20     ,+                  ,+ 4
    #       / |  with nodes:    / |
    # 10  +' ,+             1 +' ,+ 3
    #     | / |               | / |
    #  0  +---+             0 +---+ 2
    #     0  10
    mesh_args = _make_small_mesh_args()
    elem_coords = np.array([[5, 0], [5, 10]])
    node_mesh = MeshInfo(*mesh_args, location="node")
    face_mesh = MeshInfo(*mesh_args, elem_coords=elem_coords, location="face")

    # We create a grid with the following shape:
    # 20/3 +---+
    #      |   |
    # 10/3 +---+
    #      |   |
    #  0   +---+
    #      0  10
    grid_args = [ar * 2 for ar in make_grid_args(2, 3)]
    grid = GridInfo(*grid_args, center=True)

    mesh_to_grid_regridder = Regridder(
        node_mesh, grid, method=Constants.Method.BILINEAR
    )
    mesh_input = np.arange(5)
    grid_output = mesh_to_grid_regridder.regrid(mesh_input)
    # For a flat surface, we would expect the fractional part of these values
    # to be either 1/3 or 2/3. Since the actual surface lies on a sphere, and
    # due to the way ESMF calculates these values, the expected output is
    # slightly different. It's worth noting that the finer the resolution, the
    # closer these numbers are. Since the grids/meshes lie on coarse steps of
    # about 10 degrees, we can expect most cases to behave more similarly to
    # bilinear regridders on flat surfaces.
    expected_grid_output = np.array(
        [
            [0.0, 2.0],
            [0.6662902773937054, 2.6662902773105808],
            [-1, 3.333709722689418],
        ]
    )
    expected_grid_mask = np.array([[0, 0], [0, 0], [1, 0]])
    expected_grid_output = ma.array(expected_grid_output, mask=expected_grid_mask)
    assert ma.allclose(expected_grid_output, grid_output)

    grid_to_mesh_regridder = Regridder(
        grid, node_mesh, method=Constants.Method.BILINEAR
    )
    grid_input = np.array([[0, 0], [1, 0], [2, 1]])
    mesh_output = grid_to_mesh_regridder.regrid(grid_input)
    expected_mesh_output = ma.array([0.0, 1.5, 0.0, 0.5, -1], mask=[0, 0, 0, 0, 1])
    assert ma.allclose(expected_mesh_output, mesh_output)

    grid_to_face_mesh_regridder = Regridder(
        grid, face_mesh, method=Constants.Method.BILINEAR
    )
    grid_input_2 = np.array([[0, 0], [1, 0], [4, 1]])
    face_mesh_output = grid_to_face_mesh_regridder.regrid(grid_input_2)
    expected_face_mesh_output = np.array([0.0, 1.4888258584989558])
    assert ma.allclose(expected_face_mesh_output, face_mesh_output)

    def _give_extra_dims(array):
        result = np.stack([array, array + 1])
        result = np.stack([result, result + 10, result + 100])
        return result

    extra_dim_mesh_input = _give_extra_dims(mesh_input)
    extra_dim_grid_output = mesh_to_grid_regridder.regrid(extra_dim_mesh_input)
    extra_dim_expected_grid_output = _give_extra_dims(expected_grid_output)
    assert ma.allclose(extra_dim_expected_grid_output, extra_dim_grid_output)

    extra_dim_grid_input = _give_extra_dims(grid_input)
    extra_dim_mesh_output = grid_to_mesh_regridder.regrid(extra_dim_grid_input)
    extra_dim_expected_mesh_output = _give_extra_dims(expected_mesh_output)
    assert ma.allclose(extra_dim_expected_mesh_output, extra_dim_mesh_output)



"""Unit tests for :mod:`esmf_regrid.experimental.io.save_regridder`."""

import pytest

from esmf_regrid.experimental.io import _managed_var_name, save_regridder
from esmf_regrid.schemes import ESMFAreaWeightedRegridder
from esmf_regrid.tests.unit.schemes.test__mesh_to_MeshInfo import (
    _gridlike_mesh_cube,
)


def test_invalid_type(tmp_path):
    """Test that `save_regridder` raises a TypeError where appropriate."""
    invalid_obj = None
    filename = tmp_path / "regridder.nc"
    with pytest.raises(TypeError):
        save_regridder(invalid_obj, filename)


def test_var_name_preserve(tmp_path):
    """Test that `save_regridder` does not change var_ames."""
    lons = 3
    lats = 4
    src = _gridlike_mesh_cube(lons, lats)
    tgt = _gridlike_mesh_cube(lons, lats)

    DUMMY_VAR_NAME_SRC = "src_dummy_var"
    DUMMY_VAR_NAME_TGT = "tgt_dummy_var"
    for coord in src.mesh.coords():
        coord.var_name = DUMMY_VAR_NAME_SRC
    for coord in tgt.mesh.coords():
        coord.var_name = DUMMY_VAR_NAME_TGT

    rg = ESMFAreaWeightedRegridder(src, tgt)
    filename = tmp_path / "regridder.nc"
    save_regridder(rg, filename)

    for coord in src.mesh.coords():
        assert coord.var_name == DUMMY_VAR_NAME_SRC
    for coord in tgt.mesh.coords():
        assert coord.var_name == DUMMY_VAR_NAME_TGT


def test_managed_var_name():
    """Test that `_managed_var_name` changes var_names."""
    lons = 3
    lats = 4
    src = _gridlike_mesh_cube(lons, lats)
    tgt = _gridlike_mesh_cube(lons, lats)

    DUMMY_VAR_NAME_SRC = "src_dummy_var"
    DUMMY_VAR_NAME_TGT = "tgt_dummy_var"
    for coord in src.mesh.coords():
        coord.var_name = DUMMY_VAR_NAME_SRC
    for coord in tgt.mesh.coords():
        coord.var_name = DUMMY_VAR_NAME_TGT

    with _managed_var_name(src, tgt):
        for coord in src.mesh.coords():
            assert coord.var_name != DUMMY_VAR_NAME_SRC
        for coord in tgt.mesh.coords():
            assert coord.var_name != DUMMY_VAR_NAME_TGT



"""Provides load/save functions for regridders."""

from contextlib import contextmanager

import iris
from iris.coords import AuxCoord
from iris.cube import Cube, CubeList
from iris.experimental.ugrid import PARSE_UGRID_ON_LOAD
import numpy as np
import scipy.sparse

import esmf_regrid
from esmf_regrid import check_method, Constants
from esmf_regrid.experimental.unstructured_scheme import (
    GridToMeshESMFRegridder,
    MeshToGridESMFRegridder,
)
from esmf_regrid.schemes import (
    ESMFAreaWeightedRegridder,
    ESMFBilinearRegridder,
    ESMFNearestRegridder,
    GridRecord,
    MeshRecord,
)


SUPPORTED_REGRIDDERS = [
    ESMFAreaWeightedRegridder,
    ESMFBilinearRegridder,
    ESMFNearestRegridder,
    GridToMeshESMFRegridder,
    MeshToGridESMFRegridder,
]
REGRIDDER_NAME_MAP = {rg_class.__name__: rg_class for rg_class in SUPPORTED_REGRIDDERS}
SOURCE_NAME = "regridder_source_field"
SOURCE_MASK_NAME = "regridder_source_mask"
TARGET_NAME = "regridder_target_field"
TARGET_MASK_NAME = "regridder_target_mask"
WEIGHTS_NAME = "regridder_weights"
WEIGHTS_SHAPE_NAME = "weights_shape"
WEIGHTS_ROW_NAME = "weight_matrix_rows"
WEIGHTS_COL_NAME = "weight_matrix_columns"
REGRIDDER_TYPE = "regridder_type"
VERSION_ESMF = "ESMF_version"
VERSION_INITIAL = "esmf_regrid_version_on_initialise"
MDTOL = "mdtol"
METHOD = "method"
RESOLUTION = "resolution"
SOURCE_RESOLUTION = "src_resolution"
TARGET_RESOLUTION = "tgt_resolution"


def _add_mask_to_cube(mask, cube, name):
    if isinstance(mask, np.ndarray):
        mask = mask.astype(int)
        mask_coord = AuxCoord(mask, var_name=name, long_name=name)
        cube.add_aux_coord(mask_coord, list(range(cube.ndim)))


@contextmanager
def _managed_var_name(src_cube, tgt_cube):
    src_coord_names = []
    src_mesh_coords = []
    if src_cube.mesh is not None:
        src_mesh = src_cube.mesh
        src_mesh_coords = src_mesh.coords()
        for coord in src_mesh_coords:
            src_coord_names.append(coord.var_name)
    tgt_coord_names = []
    tgt_mesh_coords = []
    if tgt_cube.mesh is not None:
        tgt_mesh = tgt_cube.mesh
        tgt_mesh_coords = tgt_mesh.coords()
        for coord in tgt_mesh_coords:
            tgt_coord_names.append(coord.var_name)

    try:
        for coord in src_mesh_coords:
            coord.var_name = "_".join([SOURCE_NAME, "mesh", coord.name()])
        for coord in tgt_mesh_coords:
            coord.var_name = "_".join([TARGET_NAME, "mesh", coord.name()])
        yield None
    finally:
        for coord, var_name in zip(src_mesh_coords, src_coord_names):
            coord.var_name = var_name
        for coord, var_name in zip(tgt_mesh_coords, tgt_coord_names):
            coord.var_name = var_name


def _clean_var_names(cube):
    cube.var_name = None
    for coord in cube.coords():
        coord.var_name = None
    if cube.mesh is not None:
        cube.mesh.var_name = None
        for coord in cube.mesh.coords():
            coord.var_name = None
        for con in cube.mesh.connectivities():
            con.var_name = None


def save_regridder(rg, filename):
    """
    Save a regridder scheme instance.

    Saves any of the regridder classes, i.e.
    :class:`~esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`,
    :class:`~esmf_regrid.experimental.unstructured_scheme.MeshToGridESMFRegridder`,
    :class:`~esmf_regrid.schemes.ESMFAreaWeightedRegridder`,
    :class:`~esmf_regrid.schemes.ESMFBilinearRegridder` or
    :class:`~esmf_regrid.schemes.ESMFNearestRegridder`.
    .

    Parameters
    ----------
    rg : :class:`~esmf_regrid.schemes._ESMFRegridder`
        The regridder instance to save.
    filename : str
        The file name to save to.
    """
    regridder_type = rg.__class__.__name__

    def _standard_grid_cube(grid, name):
        if grid[0].ndim == 1:
            shape = [coord.points.size for coord in grid]
        else:
            shape = grid[0].shape
        data = np.zeros(shape)
        cube = Cube(data, var_name=name, long_name=name)
        if grid[0].ndim == 1:
            cube.add_dim_coord(grid[0], 0)
            cube.add_dim_coord(grid[1], 1)
        else:
            cube.add_aux_coord(grid[0], [0, 1])
            cube.add_aux_coord(grid[1], [0, 1])
        return cube

    def _standard_mesh_cube(mesh, location, name):
        mesh_coords = mesh.to_MeshCoords(location)
        data = np.zeros(mesh_coords[0].points.shape[0])
        cube = Cube(data, var_name=name, long_name=name)
        for coord in mesh_coords:
            cube.add_aux_coord(coord, 0)
        return cube

    if regridder_type in [
        "ESMFAreaWeightedRegridder",
        "ESMFBilinearRegridder",
        "ESMFNearestRegridder",
    ]:
        src_grid = rg._src
        if isinstance(src_grid, GridRecord):
            src_cube = _standard_grid_cube(
                (src_grid.grid_y, src_grid.grid_x), SOURCE_NAME
            )
        elif isinstance(src_grid, MeshRecord):
            src_mesh, src_location = src_grid
            src_cube = _standard_mesh_cube(src_mesh, src_location, SOURCE_NAME)
        else:
            raise ValueError("Improper type for `rg._src`.")
        _add_mask_to_cube(rg.src_mask, src_cube, SOURCE_MASK_NAME)

        tgt_grid = rg._tgt
        if isinstance(tgt_grid, GridRecord):
            tgt_cube = _standard_grid_cube(
                (tgt_grid.grid_y, tgt_grid.grid_x), TARGET_NAME
            )
        elif isinstance(tgt_grid, MeshRecord):
            tgt_mesh, tgt_location = tgt_grid
            tgt_cube = _standard_mesh_cube(tgt_mesh, tgt_location, TARGET_NAME)
        else:
            raise ValueError("Improper type for `rg._tgt`.")
        _add_mask_to_cube(rg.tgt_mask, tgt_cube, TARGET_MASK_NAME)
    elif regridder_type == "GridToMeshESMFRegridder":
        src_grid = (rg.grid_y, rg.grid_x)
        src_cube = _standard_grid_cube(src_grid, SOURCE_NAME)
        _add_mask_to_cube(rg.src_mask, src_cube, SOURCE_MASK_NAME)

        tgt_mesh = rg.mesh
        tgt_location = rg.location
        tgt_cube = _standard_mesh_cube(tgt_mesh, tgt_location, TARGET_NAME)
        _add_mask_to_cube(rg.tgt_mask, tgt_cube, TARGET_MASK_NAME)

    elif regridder_type == "MeshToGridESMFRegridder":
        src_mesh = rg.mesh
        src_location = rg.location
        src_cube = _standard_mesh_cube(src_mesh, src_location, SOURCE_NAME)
        _add_mask_to_cube(rg.src_mask, src_cube, SOURCE_MASK_NAME)

        tgt_grid = (rg.grid_y, rg.grid_x)
        tgt_cube = _standard_grid_cube(tgt_grid, TARGET_NAME)
        _add_mask_to_cube(rg.tgt_mask, tgt_cube, TARGET_MASK_NAME)
    else:
        msg = (
            f"Expected a regridder of type `GridToMeshESMFRegridder` or "
            f"`MeshToGridESMFRegridder`, got type {regridder_type}."
        )
        raise TypeError(msg)

    method = str(check_method(rg.method).name)

    if regridder_type in ["GridToMeshESMFRegridder", "MeshToGridESMFRegridder"]:
        resolution = rg.resolution
        src_resolution = None
        tgt_resolution = None
    elif regridder_type == "ESMFAreaWeightedRegridder":
        resolution = None
        src_resolution = rg.src_resolution
        tgt_resolution = rg.tgt_resolution
    else:
        resolution = None
        src_resolution = None
        tgt_resolution = None

    weight_matrix = rg.regridder.weight_matrix
    reformatted_weight_matrix = scipy.sparse.coo_matrix(weight_matrix)
    weight_data = reformatted_weight_matrix.data
    weight_rows = reformatted_weight_matrix.row
    weight_cols = reformatted_weight_matrix.col
    weight_shape = reformatted_weight_matrix.shape

    esmf_version = rg.regridder.esmf_version
    esmf_regrid_version = rg.regridder.esmf_regrid_version
    save_version = esmf_regrid.__version__

    # Currently, all schemes use the fracarea normalization.
    normalization = "fracarea"

    mdtol = rg.mdtol
    attributes = {
        "title": "iris-esmf-regrid regridding scheme",
        REGRIDDER_TYPE: regridder_type,
        VERSION_ESMF: esmf_version,
        VERSION_INITIAL: esmf_regrid_version,
        "esmf_regrid_version_on_save": save_version,
        "normalization": normalization,
        MDTOL: mdtol,
        METHOD: method,
    }
    if resolution is not None:
        attributes[RESOLUTION] = resolution
    if src_resolution is not None:
        attributes[SOURCE_RESOLUTION] = src_resolution
    if tgt_resolution is not None:
        attributes[TARGET_RESOLUTION] = tgt_resolution

    weights_cube = Cube(weight_data, var_name=WEIGHTS_NAME, long_name=WEIGHTS_NAME)
    row_coord = AuxCoord(
        weight_rows, var_name=WEIGHTS_ROW_NAME, long_name=WEIGHTS_ROW_NAME
    )
    col_coord = AuxCoord(
        weight_cols, var_name=WEIGHTS_COL_NAME, long_name=WEIGHTS_COL_NAME
    )
    weights_cube.add_aux_coord(row_coord, 0)
    weights_cube.add_aux_coord(col_coord, 0)

    weight_shape_cube = Cube(
        weight_shape,
        var_name=WEIGHTS_SHAPE_NAME,
        long_name=WEIGHTS_SHAPE_NAME,
    )

    # Save cubes while ensuring var_names do not conflict for the sake of consistency.
    with _managed_var_name(src_cube, tgt_cube):
        cube_list = CubeList([src_cube, tgt_cube, weights_cube, weight_shape_cube])

        for cube in cube_list:
            cube.attributes = attributes

        iris.fileformats.netcdf.save(cube_list, filename)


def load_regridder(filename):
    """
    Load a regridder scheme instance.

    Loads either a
    :class:`~esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder`
    or a
    :class:`~esmf_regrid.experimental.unstructured_scheme.MeshToGridESMFRegridder`.

    Parameters
    ----------
    filename : str
        The file name to load from.

    Returns
    -------
    :class:`~esmf_regrid.experimental.unstructured_scheme.GridToMeshESMFRegridder` or :class:`~esmf_regrid.experimental.unstructured_scheme.MeshToGridESMFRegridder`
    """
    with PARSE_UGRID_ON_LOAD.context():
        cubes = iris.load(filename)

    # Extract the source, target and metadata information.
    src_cube = cubes.extract_cube(SOURCE_NAME)
    _clean_var_names(src_cube)
    tgt_cube = cubes.extract_cube(TARGET_NAME)
    _clean_var_names(tgt_cube)
    weights_cube = cubes.extract_cube(WEIGHTS_NAME)
    weight_shape_cube = cubes.extract_cube(WEIGHTS_SHAPE_NAME)

    # Determine the regridder type.
    regridder_type = weights_cube.attributes[REGRIDDER_TYPE]
    assert regridder_type in REGRIDDER_NAME_MAP.keys()
    scheme = REGRIDDER_NAME_MAP[regridder_type]

    # Determine the regridding method, allowing for files created when
    # conservative regridding was the only method.
    method = getattr(
        Constants.Method, weights_cube.attributes.get(METHOD, "CONSERVATIVE")
    )

    resolution = weights_cube.attributes.get(RESOLUTION, None)
    src_resolution = weights_cube.attributes.get(SOURCE_RESOLUTION, None)
    tgt_resolution = weights_cube.attributes.get(TARGET_RESOLUTION, None)
    if resolution is not None:
        resolution = int(resolution)
    if src_resolution is not None:
        src_resolution = int(src_resolution)
    if tgt_resolution is not None:
        tgt_resolution = int(tgt_resolution)

    # Reconstruct the weight matrix.
    weight_data = weights_cube.data
    weight_rows = weights_cube.coord(WEIGHTS_ROW_NAME).points
    weight_cols = weights_cube.coord(WEIGHTS_COL_NAME).points
    weight_shape = weight_shape_cube.data
    weight_matrix = scipy.sparse.csr_matrix(
        (weight_data, (weight_rows, weight_cols)), shape=weight_shape
    )

    mdtol = weights_cube.attributes[MDTOL]

    if src_cube.coords(SOURCE_MASK_NAME):
        use_src_mask = src_cube.coord(SOURCE_MASK_NAME).points
    else:
        use_src_mask = False
    if tgt_cube.coords(TARGET_MASK_NAME):
        use_tgt_mask = tgt_cube.coord(TARGET_MASK_NAME).points
    else:
        use_tgt_mask = False

    if scheme is GridToMeshESMFRegridder:
        resolution_keyword = SOURCE_RESOLUTION
        kwargs = {resolution_keyword: resolution, "method": method, "mdtol": mdtol}
    elif scheme is MeshToGridESMFRegridder:
        resolution_keyword = TARGET_RESOLUTION
        kwargs = {resolution_keyword: resolution, "method": method, "mdtol": mdtol}
    elif scheme is ESMFAreaWeightedRegridder:
        kwargs = {
            SOURCE_RESOLUTION: src_resolution,
            TARGET_RESOLUTION: tgt_resolution,
            "mdtol": mdtol,
        }
    elif scheme is ESMFBilinearRegridder:
        kwargs = {"mdtol": mdtol}
    else:
        kwargs = {}

    regridder = scheme(
        src_cube,
        tgt_cube,
        precomputed_weights=weight_matrix,
        use_src_mask=use_src_mask,
        use_tgt_mask=use_tgt_mask,
        **kwargs,
    )

    esmf_version = weights_cube.attributes[VERSION_ESMF]
    regridder.regridder.esmf_version = esmf_version
    esmf_regrid_version = weights_cube.attributes[VERSION_INITIAL]
    regridder.regridder.esmf_regrid_version = esmf_regrid_version
    return regridder



"""Unit tests for :func:`esmf_regrid.experimental.unstructured_scheme._mesh_to_MeshInfo`."""

from iris.coords import AuxCoord
from iris.cube import Cube
from iris.experimental.ugrid import Connectivity, Mesh
import numpy as np
from numpy import ma
import scipy.sparse

from esmf_regrid.esmf_regridder import Regridder
from esmf_regrid.schemes import _mesh_to_MeshInfo


def _pyramid_topology_connectivity_array():
    """
    Generate the face_node_connectivity array for a topological pyramid.

    The mesh described is a topological pyramid in the sense that there
    exists a polygonal base (described by the indices [0, 1, 2, 3, 4])
    and all other faces are triangles connected to a single node (the node
    with index 5).
    """
    fnc_array = [
        [0, 1, 2, 3, 4],
        [1, 0, 5, -1, -1],
        [2, 1, 5, -1, -1],
        [3, 2, 5, -1, -1],
        [4, 3, 5, -1, -1],
        [0, 4, 5, -1, -1],
    ]
    fnc_mask = [
        [0, 0, 0, 0, 0],
        [0, 0, 0, 1, 1],
        [0, 0, 0, 1, 1],
        [0, 0, 0, 1, 1],
        [0, 0, 0, 1, 1],
        [0, 0, 0, 1, 1],
    ]
    fnc_ma = ma.array(fnc_array, mask=fnc_mask)
    return fnc_ma


def _example_mesh():
    """Generate a global mesh with a pentagonal pyramid topology."""
    # The base of the pyramid is the following pentagon.
    #
    # 60     0          3
    #        |  \    /  |
    # 10     |    4     |
    #        |          |
    #        |          |
    # -60    1----------2
    #
    #      120   180  -120
    #
    # The point of the pyramid is at the coordinate (0, 0).
    # The geometry is designed so that a valid ESMF object is only produced when
    # the orientation is correct (the face nodes are visited in an anticlockwise
    # order). This sensitivity is due to the base of the pyramid being convex.

    # Generate face_node_connectivity (fnc).
    fnc_ma = _pyramid_topology_connectivity_array()
    fnc = Connectivity(
        fnc_ma,
        cf_role="face_node_connectivity",
        start_index=0,
    )
    lon_values = [120, 120, -120, -120, 180, 0]
    lat_values = [60, -60, -60, 60, 10, 0]
    lons = AuxCoord(lon_values, standard_name="longitude")
    lats = AuxCoord(lat_values, standard_name="latitude")
    mesh = Mesh(2, ((lons, "x"), (lats, "y")), fnc)
    return mesh


def _gridlike_mesh(n_lons, n_lats, nsi=0):
    """
    Generate a global mesh with geometry similar to a rectilinear grid.

    The resulting mesh will have n_lons cells spanning its longitudes and
    n_lats cells spanning its latitudes for a total of (n_lons * n_lats) cells.
    Note that the cells neighbouring the poles will actually be triangular while
    the rest of the cells will be rectangular.
    """
    # Arrange the indices of the non-pole nodes in an array representative of their
    # latitude/longitude.
    fnc_template = np.arange((n_lats - 1) * n_lons).reshape(n_lats - 1, n_lons) + 1
    fnc_array = np.empty([n_lats, n_lons, 4])
    # Assign points in an anticlockwise orientation. From the 0 node to 1
    # longitude increases, then from 1 to 2 latitude increases, from 2 to 3
    # longitude decreases and from 3 to 0 latitude decreases.
    fnc_array[1:, :, 0] = fnc_template
    fnc_array[1:, :, 1] = np.roll(fnc_template, -1, 1)
    fnc_array[:-1, :, 2] = np.roll(fnc_template, -1, 1)
    fnc_array[:-1, :, 3] = fnc_template
    # Define the poles as single points. Note that all the cells adjacent to these
    # nodes will be topologically triangular with the pole node repeated. One of
    # these repeated pole node references will eventually be masked.
    fnc_array[0, :, :2] = 0
    num_nodes = fnc_template.max()
    fnc_array[-1, :, 2:] = num_nodes + 1
    # By convention, node references to be masked should be last in the list.
    # Since one of the pole node references will end up masked, this should be
    # moved to the end of the list of nodes.
    fnc_array[0, :, :] = np.roll(fnc_array[0, :, :], -1, -1)

    # One of the two references to the pole node are defined to be masked.
    fnc_mask = np.zeros_like(fnc_array)
    fnc_mask[0, :, -1] = 1
    fnc_mask[-1, :, -1] = 1
    fnc_ma = ma.array(fnc_array, mask=fnc_mask, dtype=int)

    # The face node connectivity is flattened to the correct dimensionality.
    fnc_ma = fnc_ma.reshape([-1, 4])

    # Describe the edge node connectivity.
    # There are n_lats * n_lons vertical edges and (n_lats - 1) * n_lons horizontal
    # edges which we arrange into an array for convenience of calculation.
    enc_array = np.empty([(n_lats * 2) - 1, n_lons, 2], dtype=int)
    # The vertical edges make up enc_array[:n_lats].
    enc_array[1:n_lats, :, 0] = fnc_template
    enc_array[: n_lats - 1, :, 1] = fnc_template
    enc_array[0, :, 0] = 0
    # The horizontal edges make up enc_array[n_lats:].
    enc_array[n_lats - 1, :, 1] = num_nodes + 1
    enc_array[n_lats:, :, 0] = fnc_template
    enc_array[n_lats:, :, 1] = np.roll(fnc_template, -1, 1)
    # The array is flattened to its proper shape of (N, 2).
    enc_array = enc_array.reshape([-1, 2])

    # Latitude and longitude values are set.
    lat_values = np.linspace(-90, 90, n_lats + 1)
    lon_values = np.linspace(-180, 180, n_lons, endpoint=False)
    # Latitude values are broadcast to arrays with the same shape as the face node
    # connectivity node references in fnc_template.
    lon_array, lat_array = np.meshgrid(lon_values, lat_values[1:-1])
    node_lats = np.empty(num_nodes + 2)
    # Note that fnc_template is created by reshaping a list of  indices. These
    # indices refer to node_lats and node_lons which are generated by reshaping
    # lat_array and lon_array. Because of the way reshaping preserves order, there
    # is a correspondance between an index in a particular position fnc_template
    # and the latitude and longitude described by lat_array and lon_array in the
    # same position.
    node_lats[1:-1] = lat_array.reshape([-1])
    # Define the latitude and longitude of the poles.
    node_lats[0] = lat_values[0]
    node_lats[-1] = lat_values[-1]
    node_lons = np.empty(num_nodes + 2)
    node_lons[1:-1] = lon_array.reshape([-1])
    node_lons[0] = 0
    node_lons[-1] = 0

    # Center Latitude and Longitude values are set.
    lon_centers = np.linspace(-180, 180, (2 * n_lons) + 1)[1::2]
    lat_centers = np.linspace(-90, 90, (2 * n_lats) + 1)[1::2]
    lon_center_array, lat_center_array = np.meshgrid(lon_centers, lat_centers)
    face_lons = lon_center_array.flatten()
    face_lats = lat_center_array.flatten()

    # Translate the mesh information into iris objects.
    fnc = Connectivity(
        fnc_ma + nsi,
        cf_role="face_node_connectivity",
        start_index=nsi,
    )
    enc = Connectivity(
        enc_array,
        cf_role="edge_node_connectivity",
        start_index=0,
    )
    lons = AuxCoord(node_lons, standard_name="longitude")
    lats = AuxCoord(node_lats, standard_name="latitude")
    mesh = Mesh(2, ((lons, "x"), (lats, "y")), [fnc, enc])

    # In order to add a mesh to a cube, face locations must be added.
    face_lon_coord = AuxCoord(face_lons, standard_name="longitude")
    face_lat_coord = AuxCoord(face_lats, standard_name="latitude")

    # Add dummy edge coords.
    dummy_points = np.zeros(enc_array.shape[0])
    edge_lon_coord = AuxCoord(dummy_points, standard_name="longitude")
    edge_lat_coord = AuxCoord(dummy_points, standard_name="latitude")

    mesh.add_coords(
        face_x=face_lon_coord,
        face_y=face_lat_coord,
        edge_x=edge_lon_coord,
        edge_y=edge_lat_coord,
    )
    mesh.long_name = "example mesh"
    return mesh


def _gridlike_mesh_cube(n_lons, n_lats, location="face", nsi=0):
    mesh = _gridlike_mesh(n_lons, n_lats, nsi=nsi)
    mesh_coord_x, mesh_coord_y = mesh.to_MeshCoords(location)
    data = np.zeros_like(mesh_coord_x.points)
    cube = Cube(data)
    cube.add_aux_coord(mesh_coord_x, 0)
    cube.add_aux_coord(mesh_coord_y, 0)
    return cube


def test__mesh_to_MeshInfo():
    """Basic test for :func:`esmf_regrid.experimental.unstructured_scheme._mesh_to_MeshInfo`."""
    mesh = _example_mesh()
    meshinfo = _mesh_to_MeshInfo(mesh, location="face")

    expected_nodes = np.array(
        [
            [120, 60],
            [120, -60],
            [-120, -60],
            [-120, 60],
            [180, 10],
            [0, 0],
        ]
    )
    assert np.array_equal(expected_nodes, meshinfo.node_coords)

    expected_connectivity = _pyramid_topology_connectivity_array()
    assert np.array_equal(expected_connectivity, meshinfo.fnc)

    expected_start_index = 0
    assert expected_start_index == meshinfo.esi


def test_anticlockwise_validity():
    """Test validity of objects derived from Mesh objects with anticlockwise orientation."""
    mesh = _example_mesh()
    meshinfo = _mesh_to_MeshInfo(mesh, location="face")

    # Ensure conversion to ESMF works without error.
    _ = meshinfo.make_esmf_field()

    # The following test ensures there are no overlapping cells.
    # This catches geometric/topological abnormalities that would arise from,
    # for example: switching lat/lon values, using euclidean coords vs spherical.
    rg = Regridder(meshinfo, meshinfo)
    expected_weights = scipy.sparse.identity(meshinfo.size)
    assert np.allclose(expected_weights.todense(), rg.weight_matrix.todense())


def test_large_mesh_validity():
    """Test validity of objects derived from a large gridlike Mesh."""
    mesh = _gridlike_mesh(40, 20)
    meshinfo = _mesh_to_MeshInfo(mesh, location="face")

    # Ensure conversion to ESMF works without error.
    _ = meshinfo.make_esmf_field()

    # The following test ensures there are no overlapping cells.
    # This catches geometric/topological abnormalities that would arise from,
    # for example: switching lat/lon values, using euclidean coords vs spherical.
    rg = Regridder(meshinfo, meshinfo)
    expected_weights = scipy.sparse.identity(meshinfo.size)
    assert np.allclose(expected_weights.todense(), rg.weight_matrix.todense())



"""Unit tests for :class:`esmf_regrid.schemes.ESMFBilinearRegridder`."""

from cf_units import Unit
import numpy as np
import pytest

from esmf_regrid.schemes import ESMFBilinearRegridder
from esmf_regrid.tests.unit.schemes.test__cube_to_GridInfo import (
    _curvilinear_cube,
    _grid_cube,
)


def test_dim_switching():
    """
    Test calling of :class:`esmf_regrid.schemes.ESMFBilinearRegridder`.

    Checks that the regridder accepts a cube with dimensions in a different
    order than the cube which initialised it. Checks that dimension order is
    inherited from the cube in the calling function in both cases.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    tgt = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    regridder = ESMFBilinearRegridder(src, tgt)
    unswitched_result = regridder(src)

    src_switched = src.copy()
    src_switched.transpose()
    switched_result = regridder(src_switched)

    assert unswitched_result.coord(dimensions=(0,)).standard_name == "latitude"
    assert unswitched_result.coord(dimensions=(1,)).standard_name == "longitude"
    assert switched_result.coord(dimensions=(0,)).standard_name == "longitude"
    assert switched_result.coord(dimensions=(1,)).standard_name == "latitude"


def test_differing_grids():
    """
    Test calling of :class:`esmf_regrid.schemes.ESMFBilinearRegridder`.

    Checks that the regridder raises an error when given a cube with a different
    grid to the one it was initialised with.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src_init = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    tgt = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    n_lons_dif = 7
    src_dif_coord = _grid_cube(
        n_lons_dif, n_lats, lon_bounds, lat_bounds, circular=True
    )
    src_dif_circ = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=False)

    regridder = ESMFBilinearRegridder(src_init, tgt)

    msg = "The given cube is not defined on the same source grid as this regridder."
    with pytest.raises(ValueError, match=msg):
        _ = regridder(src_dif_coord)
    with pytest.raises(ValueError, match=msg):
        _ = regridder(src_dif_circ)


def test_invalid_mdtol():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFBilinearRegridder`.

    Checks that an error is raised when mdtol is out of range.
    """
    n_lons = 6
    n_lats = 5
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    src = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)
    tgt = _grid_cube(n_lons, n_lats, lon_bounds, lat_bounds, circular=True)

    match = "Value for mdtol must be in range 0 - 1, got "
    with pytest.raises(ValueError, match=match):
        _ = ESMFBilinearRegridder(src, tgt, mdtol=2)
    with pytest.raises(ValueError, match=match):
        _ = ESMFBilinearRegridder(src, tgt, mdtol=-1)


def test_curvilinear_equivalence():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFBilinearRegridder`.

    Checks that equivalent curvilinear and rectilinear coordinates give the same
    results.
    """
    n_lons_src = 6
    n_lons_tgt = 3
    n_lats_src = 4
    n_lats_tgt = 2
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    grid_src = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
    grid_tgt = _grid_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds, circular=True)
    curv_src = _curvilinear_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds)
    curv_tgt = _curvilinear_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds)

    grid_to_grid = ESMFBilinearRegridder(grid_src, grid_tgt)
    grid_to_curv = ESMFBilinearRegridder(grid_src, curv_tgt)
    curv_to_grid = ESMFBilinearRegridder(curv_src, grid_tgt)
    curv_to_curv = ESMFBilinearRegridder(curv_src, curv_tgt)

    def extract_weights(regridder):
        return regridder.regridder.weight_matrix.todense()

    for regridder in [grid_to_curv, curv_to_grid, curv_to_curv]:
        assert np.allclose(extract_weights(grid_to_grid), extract_weights(regridder))


def test_curvilinear_and_rectilinear():
    """
    Test :class:`esmf_regrid.schemes.ESMFBilinearRegridder`.

    Checks that a cube with both curvilinear and rectilinear coords still works.
    Checks that the DimCoords have priority over AuxCoords.
    """
    n_lons_src = 6
    n_lons_tgt = 3
    n_lats_src = 4
    n_lats_tgt = 2
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    grid_src = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
    grid_tgt = _grid_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds, circular=True)
    curv_src = _curvilinear_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds)
    curv_tgt = _curvilinear_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds)

    src = curv_src.copy()
    grid_lat_src = grid_src.coord("latitude")
    grid_lat_src.standard_name = "grid_latitude"
    src.add_dim_coord(grid_lat_src, 0)
    grid_lon_src = grid_src.coord("longitude")
    grid_lon_src.standard_name = "grid_longitude"
    src.add_dim_coord(grid_lon_src, 1)

    tgt = curv_tgt.copy()
    grid_lat_tgt = grid_tgt.coord("latitude")
    grid_lat_tgt.standard_name = "grid_latitude"
    tgt.add_dim_coord(grid_lat_tgt, 0)
    grid_lon_tgt = grid_tgt.coord("longitude")
    grid_lon_tgt.standard_name = "grid_longitude"
    tgt.add_dim_coord(grid_lon_tgt, 1)

    # Change the AuxCoords to check that the DimCoords have priority.
    src.coord("latitude").bounds[:] = 0
    src.coord("longitude").bounds[:] = 0
    tgt.coord("latitude").bounds[:] = 0
    tgt.coord("longitude").bounds[:] = 0

    # Ensure the source data is all ones.
    src.data[:] = 1

    rg = ESMFBilinearRegridder(src, tgt)
    result = rg(src)

    expected = grid_tgt.copy()
    # If the aux coords had been prioritised, expected.data would be a fully masked array.
    expected.data[:] = 1
    assert expected == result
    assert not np.ma.is_masked(result)


def test_unit_equivalence():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFBilinearRegridder`.

    Checks that equivalent coordinates in degrees and radians give the same results.
    """
    n_lons_src = 6
    n_lons_tgt = 3
    n_lats_src = 4
    n_lats_tgt = 2
    lon_bounds = (-180, 180)
    lat_bounds = (-90, 90)
    lon_rad_bounds = (-np.pi, np.pi)
    lat_rad_bounds = (-np.pi / 2, np.pi / 2)

    def rad_coords(cube):
        cube.coord("latitude").units = Unit("radians")
        cube.coord("longitude").units = Unit("radians")

    grid_src = _grid_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds, circular=True)
    grid_src_rad = _grid_cube(
        n_lons_src, n_lats_src, lon_rad_bounds, lat_rad_bounds, circular=True
    )
    rad_coords(grid_src_rad)
    grid_tgt = _grid_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds, circular=True)
    grid_tgt_rad = _grid_cube(
        n_lons_tgt, n_lats_tgt, lon_rad_bounds, lat_rad_bounds, circular=True
    )
    rad_coords(grid_tgt_rad)
    curv_src = _curvilinear_cube(n_lons_src, n_lats_src, lon_bounds, lat_bounds)
    curv_src_rad = _curvilinear_cube(
        n_lons_src, n_lats_src, lon_rad_bounds, lat_rad_bounds
    )
    rad_coords(curv_src_rad)
    curv_tgt = _curvilinear_cube(n_lons_tgt, n_lats_tgt, lon_bounds, lat_bounds)
    curv_tgt_rad = _curvilinear_cube(
        n_lons_tgt, n_lats_tgt, lon_rad_bounds, lat_rad_bounds
    )
    rad_coords(curv_tgt_rad)

    grid_to_grid = ESMFBilinearRegridder(grid_src, grid_tgt)
    grid_rad_to_grid = ESMFBilinearRegridder(grid_src_rad, grid_tgt)
    grid_rad_to_curv = ESMFBilinearRegridder(grid_src_rad, curv_tgt)
    curv_to_grid_rad = ESMFBilinearRegridder(curv_src, grid_tgt_rad)
    curv_rad_to_grid = ESMFBilinearRegridder(curv_src_rad, grid_tgt)
    curv_to_curv_rad = ESMFBilinearRegridder(curv_src, curv_tgt_rad)

    def extract_weights(regridder):
        return regridder.regridder.weight_matrix.todense()

    for regridder in [
        grid_rad_to_grid,
        grid_rad_to_curv,
        curv_to_grid_rad,
        curv_rad_to_grid,
        curv_to_curv_rad,
    ]:
        assert np.allclose(extract_weights(grid_to_grid), extract_weights(regridder))


def test_masks():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFBilinearRegridder`.

    Checks that the `use_src_mask` and `use_tgt_mask` keywords work properly.
    """
    src = _curvilinear_cube(7, 6, [-180, 180], [-90, 90])
    tgt = _curvilinear_cube(6, 7, [-180, 180], [-90, 90])

    # Make src and tgt discontiguous at (0, 0)
    src_mask = np.zeros([6, 7], dtype=bool)
    src_mask[0, 0] = True
    src.data = np.ma.array(src.data, mask=src_mask)
    src_discontiguous = src.copy()
    src_discontiguous.coord("latitude").bounds[0, 0] = 0
    src_discontiguous.coord("longitude").bounds[0, 0] = 0

    tgt_mask = np.zeros([7, 6], dtype=bool)
    tgt_mask[0, 0] = True
    tgt.data = np.ma.array(tgt.data, mask=tgt_mask)
    tgt_discontiguous = tgt.copy()
    tgt_discontiguous.coord("latitude").bounds[0, 0] = 0
    tgt_discontiguous.coord("longitude").bounds[0, 0] = 0

    rg_src_masked = ESMFBilinearRegridder(src_discontiguous, tgt, use_src_mask=True)
    rg_tgt_masked = ESMFBilinearRegridder(src, tgt_discontiguous, use_tgt_mask=True)
    rg_unmasked = ESMFBilinearRegridder(src, tgt)

    weights_src_masked = rg_src_masked.regridder.weight_matrix
    weights_tgt_masked = rg_tgt_masked.regridder.weight_matrix
    weights_unmasked = rg_unmasked.regridder.weight_matrix

    # Check there are no weights associated with the masked point.
    assert weights_src_masked[:, 0].nnz == 0
    assert weights_tgt_masked[0].nnz == 0

    # Check all other weights are correct.
    assert np.allclose(weights_src_masked[2:].todense(), weights_unmasked[2:].todense())
    assert np.allclose(weights_tgt_masked[2:].todense(), weights_unmasked[2:].todense())


def test_regrid_data():
    """
    Test initialisation of :class:`esmf_regrid.schemes.ESMFBilinearRegridder`.

    Checks that regridding mathematics behaves in an expected way.
    """
    # Create two similar grids so that source data and expected result
    # data ought to look similar by visual inspection.
    src = _grid_cube(5, 4, [-180, 180], [-90, 90], circular=True)
    tgt = _grid_cube(4, 5, [-180, 180], [-90, 90], circular=True)

    src_data = np.arange(20).reshape([4, 5])
    src.data = src_data
    rg = ESMFBilinearRegridder(src, tgt)

    expected_data = np.array(
        [
            [
                0.3844461499074716,
                1.4148972061736933,
                2.585102793826307,
                3.615553850092528,
            ],
            [
                3.886915086072201,
                5.423003230276512,
                6.641000710423149,
                7.588216702776281,
            ],
            [
                7.649349191647961,
                8.891001259926682,
                10.108998740073318,
                11.35065080835204,
            ],
            [
                11.411783297223723,
                12.358999289576854,
                13.576996769723491,
                15.1130849139278,
            ],
            [
                15.384446149907474,
                16.414897206173688,
                17.58510279382631,
                18.615553850092528,
            ],
        ]
    )
    result = rg(src)
    np.testing.assert_allclose(expected_data, result.data)



"""Provides ESMF representations of grids/UGRID meshes and a modified regridder."""

import numpy as np
from numpy import ma
import scipy.sparse

import esmf_regrid
from esmf_regrid import check_method, check_norm, Constants
from . import esmpy
from ._esmf_sdo import GridInfo, RefinedGridInfo

__all__ = [
    "GridInfo",
    "RefinedGridInfo",
    "Regridder",
]


def _get_regrid_weights_dict(src_field, tgt_field, regrid_method):
    # The value, in array form, that ESMF should treat as an affirmative mask.
    expected_mask = np.array([True])
    regridder = esmpy.Regrid(
        src_field,
        tgt_field,
        ignore_degenerate=True,
        regrid_method=regrid_method,
        unmapped_action=esmpy.UnmappedAction.IGNORE,
        # Choosing the norm_type DSTAREA allows for mdtol type operations
        # to be performed using the weights information later on.
        norm_type=esmpy.NormType.DSTAREA,
        src_mask_values=expected_mask,
        dst_mask_values=expected_mask,
        factors=True,
    )
    # Without specifying deep_copy=true, the information in weights_dict
    # would be corrupted when the ESMF regridder is destoyed.
    weights_dict = regridder.get_weights_dict(deep_copy=True)
    # The weights_dict contains all the information needed for regridding,
    # the ESMF objects can be safely removed.
    regridder.destroy()
    return weights_dict


def _weights_dict_to_sparse_array(weights, shape, index_offsets):
    matrix = scipy.sparse.csr_matrix(
        (
            weights["weights"],
            (
                weights["row_dst"] - index_offsets[0],
                weights["col_src"] - index_offsets[1],
            ),
        ),
        shape=shape,
    )
    return matrix


class Regridder:
    """Regridder for directly interfacing with :mod:`esmpy`."""

    def __init__(
        self, src, tgt, method=Constants.Method.CONSERVATIVE, precomputed_weights=None
    ):
        """
        Create a regridder from descriptions of horizontal grids/meshes.

        Weights will be calculated using :mod:`esmpy` and stored as a
        :class:`scipy.sparse.csr_matrix`
        for use in regridding. If precomputed weights are provided,
        these will be used instead of calculating via :mod:`esmpy`.

        Parameters
        ----------
        src : :class:`~esmf_regrid.experimental.unstructured_regrid.MeshInfo` or :class:`GridInfo`
            Describes the source mesh/grid.
            Data supplied to this regridder should be in a :class:`numpy.ndarray`
            whose shape is compatible with ``src``.
        tgt : :class:`~esmf_regrid.experimental.unstructured_regrid.MeshInfo` or :class:`GridInfo`
            Describes the target mesh/grid.
            Data output by this regridder will be a :class:`numpy.ndarray` whose
            shape is compatible with ``tgt``.
        method : :class:`Constants.Method`
            The method to be used to calculate weights.
        precomputed_weights : :class:`scipy.sparse.spmatrix`, optional
            If ``None``, :mod:`esmpy` will be used to
            calculate regridding weights. Otherwise, :mod:`esmpy` will be bypassed
            and ``precomputed_weights`` will be used as the regridding weights.
        """
        self.src = src
        self.tgt = tgt
        # type checks method
        self.method = check_method(method)

        self.esmf_regrid_version = esmf_regrid.__version__
        if precomputed_weights is None:
            self.esmf_version = esmpy.__version__
            weights_dict = _get_regrid_weights_dict(
                src.make_esmf_field(),
                tgt.make_esmf_field(),
                regrid_method=method.value,
            )
            self.weight_matrix = _weights_dict_to_sparse_array(
                weights_dict,
                (self.tgt._refined_size, self.src._refined_size),
                (self.tgt.index_offset, self.src.index_offset),
            )
            if isinstance(tgt, RefinedGridInfo):
                # At this point, the weight matrix represents more target points than
                # tgt respresents. In order to collapse these points, we collapse the
                # weights matrix by the appropriate matrix multiplication.
                self.weight_matrix = (
                    tgt._collapse_weights(is_tgt=True) @ self.weight_matrix
                )
            if isinstance(src, RefinedGridInfo):
                # At this point, the weight matrix represents more source points than
                # src respresents. In order to collapse these points, we collapse the
                # weights matrix by the appropriate matrix multiplication.
                self.weight_matrix = self.weight_matrix @ src._collapse_weights(
                    is_tgt=False
                )
        else:
            if not scipy.sparse.isspmatrix(precomputed_weights):
                raise ValueError(
                    "Precomputed weights must be given as a sparse matrix."
                )
            if precomputed_weights.shape != (self.tgt.size, self.src.size):
                msg = "Expected precomputed weights to have shape {}, got shape {} instead."
                raise ValueError(
                    msg.format(
                        (self.tgt.size, self.src.size),
                        precomputed_weights.shape,
                    )
                )
            self.esmf_version = None
            self.weight_matrix = precomputed_weights

    def _out_dtype(self, in_dtype):
        """Return the expected output dtype for a given input dtype."""
        weight_dtype = self.weight_matrix.dtype
        out_dtype = (np.ones(1, dtype=in_dtype) * np.ones(1, dtype=weight_dtype)).dtype
        return out_dtype

    def regrid(self, src_array, norm_type=Constants.NormType.FRACAREA, mdtol=1):
        """
        Perform regridding on an array of data.

        Parameters
        ----------
        src_array : :obj:`~numpy.typing.ArrayLike`
            Array whose shape is compatible with ``self.src``
        norm_type : :class:`Constants.NormType`
            Either ``Constants.NormType.FRACAREA`` or ``Constants.NormType.DSTAREA``.
            Determines the type of normalisation applied to the weights.
        mdtol : float, default=1
            A number between 0 and 1 describing the missing data tolerance.
            Depending on the value of ``mdtol``, if a cell in the target grid is not
            sufficiently covered by unmasked cells of the source grid, then it will
            be masked. ``mdtol=1`` means that only target cells which are not
            covered at all will be masked, ``mdtol=0`` means that all target
            cells that are not entirely covered will be masked, and ``mdtol=0.5``
            means that all target cells that are less than half covered will
            be masked.

        Returns
        -------
        :obj:`~numpy.typing.ArrayLike`
            An array whose shape is compatible with ``self.tgt``.

        """
        # Sets default value, as this can't be done with class attributes within method call
        norm_type = check_norm(norm_type)

        array_shape = src_array.shape
        main_shape = array_shape[-self.src.dims :]
        if main_shape != self.src.shape:
            raise ValueError(
                f"Expected an array whose shape ends in {self.src.shape}, "
                f"got an array with shape ending in {main_shape}."
            )
        extra_shape = array_shape[: -self.src.dims]
        extra_size = max(1, np.prod(extra_shape))
        src_inverted_mask = self.src._array_to_matrix(~ma.getmaskarray(src_array))
        weight_sums = self.weight_matrix @ src_inverted_mask
        out_dtype = self._out_dtype(src_array.dtype)
        # Set the minimum mdtol to be slightly higher than 0 to account for rounding
        # errors.
        mdtol = max(mdtol, 1e-8)
        tgt_mask = weight_sums > 1 - mdtol
        masked_weight_sums = weight_sums * tgt_mask
        normalisations = np.ones([self.tgt.size, extra_size], dtype=out_dtype)
        if norm_type == Constants.NormType.FRACAREA:
            normalisations[tgt_mask] /= masked_weight_sums[tgt_mask]
        elif norm_type == Constants.NormType.DSTAREA:
            pass
        normalisations = ma.array(normalisations, mask=np.logical_not(tgt_mask))

        flat_src = self.src._array_to_matrix(ma.filled(src_array, 0.0))
        flat_tgt = self.weight_matrix @ flat_src
        flat_tgt = flat_tgt * normalisations
        tgt_array = self.tgt._matrix_to_array(flat_tgt, extra_shape)
        return tgt_array



"""Provides an iris interface for regridding."""

from collections import namedtuple
import copy

from cf_units import Unit
import dask.array as da
import iris.coords
import iris.cube
from iris.exceptions import CoordinateNotFoundError
from iris.experimental.ugrid import Mesh
import numpy as np

from esmf_regrid import check_method, Constants
from esmf_regrid.esmf_regridder import GridInfo, RefinedGridInfo, Regridder
from esmf_regrid.experimental.unstructured_regrid import MeshInfo

__all__ = [
    "ESMFAreaWeighted",
    "ESMFAreaWeightedRegridder",
    "ESMFBilinear",
    "ESMFBilinearRegridder",
    "ESMFNearest",
    "ESMFNearestRegridder",
    "regrid_rectilinear_to_rectilinear",
]


def _get_coord(cube, axis):
    try:
        coord = cube.coord(axis=axis, dim_coords=True)
    except CoordinateNotFoundError:
        coord = cube.coord(axis=axis)
    return coord


def _get_mask(cube_or_mesh, use_mask=True):
    if use_mask is False:
        result = None
    elif use_mask is True:
        if isinstance(cube_or_mesh, Mesh):
            result = None
        else:
            cube = cube_or_mesh
            src_x, src_y = (_get_coord(cube, "x"), _get_coord(cube, "y"))

            horizontal_dims = set(cube.coord_dims(src_x)) | set(cube.coord_dims(src_y))
            other_dims = tuple(set(range(cube.ndim)) - horizontal_dims)

            # Find a representative slice of data that spans both horizontal coords.
            if cube.coord_dims(src_x) == cube.coord_dims(src_y):
                slices = cube.slices([src_x])
            else:
                slices = cube.slices([src_x, src_y])
            data = next(slices).data
            if np.ma.is_masked(data):
                # Check that the mask is constant along all other dimensions.
                full_mask = da.ma.getmaskarray(cube.core_data())
                if not np.array_equal(
                    da.all(full_mask, axis=other_dims).compute(),
                    da.any(full_mask, axis=other_dims).compute(),
                ):
                    raise ValueError(
                        "The mask derived from the cube is not constant over non-horizontal dimensions."
                        "Consider passing in an explicit mask instead."
                    )
                mask = np.ma.getmaskarray(data)
                # Due to structural reasons, the mask should be transposed for curvilinear grids.
                if cube.coord_dims(src_x) != cube.coord_dims(src_y):
                    mask = mask.T
            else:
                mask = None
            result = mask
    else:
        result = use_mask
    return result


def _contiguous_masked(bounds, mask):
    """
    Return the (N+1, M+1) vertices for 2D coordinate bounds of shape (N, M, 4).

    Assumes the bounds are contiguous outside of the mask. As long as the only
    discontiguities are associated with masked points, the returned vertices will
    represent every bound associated with an unmasked point. This function is
    designed to replicate the behaviour of iris.Coord.contiguous_bounds() for such
    bounds. For each vertex in the returned array there are up to four choices of
    bounds to derive from. Bounds associated with umasked points will be prioritised
    in this choice.

    For example, suppose we have a masked array:

    # 0 0 0 0
    # 0 - - 0
    # 0 - - 0
    # 0 0 0 0

    Now the indices of the final bounds dimension correspond to positions on the
    vertex array. For a bound whose first indixes are (i, j) the corresponding
    position in the vertex array of the four final indices are as follows:
    0=(j, i), 1=(j, i+1), 2=(j+1, i+1), 3=(j+1, i)
    The indices of the bounds which the final array will derive from are as follows:

    # (0, 0, 0) (0, 1, 0) (0, 2, 0) (0, 3, 0) (0, 3, 1)
    # (1, 0, 0) (1, 0, 1) (0, 2, 3) (1, 3, 0) (1, 3, 1)
    # (2, 0, 0) (2, 0, 1) (1, 1, 2) (2, 3, 0) (2, 3, 1)
    # (3, 0, 0) (3, 1, 0) (3, 2, 0) (3, 3, 0) (3, 3, 1)
    # (3, 0, 3) (3, 1, 3) (3, 3, 3) (3, 3, 3) (3, 3, 2)

    Note that only the center bound derives from a masked cell as there are no
    unmasked points to derive from.
    """
    mask = np.array(mask, dtype=int)

    # Construct a blank template array in the shape of the output.
    blank_template = np.zeros([size + 1 for size in mask.shape], dtype=int)

    # Define the slices of the resulting vertex array which derive from the
    # bounds in index 0 to 3.
    slice_0 = np.s_[:-1, :-1]
    slice_1 = np.s_[:-1, 1:]
    slice_2 = np.s_[1:, 1:]
    slice_3 = np.s_[1:, :-1]

    # Define the bounds which will derive from the bounds in index 0.
    filter_0 = blank_template.copy()
    filter_0[slice_0] = 1 - mask
    # Ensure the corner points are covered appropriately.
    filter_0[0, 0] = 1

    # Define the bounds which will derive from the bounds in index 1.
    filter_1 = blank_template.copy()
    filter_1[slice_1] = 1 - mask
    # Ensure the corner and edge bounds are covered appropriately.
    filter_1[0, 1:] = 1
    # Do not cover any points already covered.
    filter_1 = filter_1 * (1 - filter_0)

    # Define the bounds which will derive from the bounds in index 3.
    filter_3 = blank_template.copy()
    filter_3[slice_3] = 1 - mask
    # Ensure the corner and edge bounds are covered appropriately.
    filter_3[1:, 0] = 1
    # Do not cover any points already covered.
    filter_3 = filter_3 * (1 - filter_0 - filter_1)

    # The bounds deriving from index 2 will be all those not already covered.
    filter_2 = 1 - filter_0 - filter_1 - filter_3

    # Copy the bounds information over into the appropriate places.
    contiguous_bounds = blank_template.astype(bounds.dtype)
    contiguous_bounds[slice_0] += bounds[:, :, 0] * filter_0[slice_0]
    contiguous_bounds[slice_1] += bounds[:, :, 1] * filter_1[slice_1]
    contiguous_bounds[slice_2] += bounds[:, :, 2] * filter_2[slice_2]
    contiguous_bounds[slice_3] += bounds[:, :, 3] * filter_3[slice_3]
    return contiguous_bounds


def _cube_to_GridInfo(cube, center=False, resolution=None, mask=None):
    # This is a simplified version of an equivalent function/method in PR #26.
    # It is anticipated that this function will be replaced by the one in PR #26.
    #
    # Returns a GridInfo object describing the horizontal grid of the cube.
    # This may be inherited from code written for the rectilinear regridding scheme.
    lon, lat = _get_coord(cube, "x"), _get_coord(cube, "y")
    #  Checks may cover units, coord systems (e.g. rotated pole), contiguous bounds.
    if cube.coord_system() is None:
        crs = None
    else:
        crs = cube.coord_system().as_cartopy_crs()
    londim, latdim = len(lon.shape), len(lat.shape)
    assert londim == latdim
    assert londim in (1, 2)
    if londim == 1:
        # Ensure coords come from a proper grid.
        assert isinstance(lon, iris.coords.DimCoord)
        assert isinstance(lat, iris.coords.DimCoord)
        circular = lon.circular
        lon_bound_array = lon.contiguous_bounds()
        lat_bound_array = lat.contiguous_bounds()
        # TODO: perform checks on lat/lon.
    elif londim == 2:
        assert cube.coord_dims(lon) == cube.coord_dims(lat)
        if not np.any(mask):
            assert lon.is_contiguous()
            assert lat.is_contiguous()
            lon_bound_array = lon.contiguous_bounds()
            lat_bound_array = lat.contiguous_bounds()
        else:
            lon_bound_array = _contiguous_masked(lon.bounds, mask)
            lat_bound_array = _contiguous_masked(lat.bounds, mask)
        # 2D coords must be AuxCoords, which do not have a circular attribute.
        circular = False
    lon_points = lon.points
    lat_points = lat.points
    if crs is None:
        lon_bound_array = lon.units.convert(lon_bound_array, Unit("degrees"))
        lat_bound_array = lat.units.convert(lat_bound_array, Unit("degrees"))
        lon_points = lon.units.convert(lon_points, Unit("degrees"))
        lat_points = lon.units.convert(lat_points, Unit("degrees"))
    if resolution is None:
        grid_info = GridInfo(
            lon_points,
            lat_points,
            lon_bound_array,
            lat_bound_array,
            crs=crs,
            circular=circular,
            center=center,
            mask=mask,
        )
    else:
        grid_info = RefinedGridInfo(
            lon_bound_array,
            lat_bound_array,
            crs=crs,
            resolution=resolution,
            mask=mask,
        )
    return grid_info


def _mesh_to_MeshInfo(mesh, location, mask=None):
    # Returns a MeshInfo object describing the mesh of the cube.
    assert mesh.topology_dimension == 2
    if None in mesh.face_coords:
        elem_coords = None
    else:
        elem_coords = np.stack([coord.points for coord in mesh.face_coords], axis=-1)
    meshinfo = MeshInfo(
        np.stack([coord.points for coord in mesh.node_coords], axis=-1),
        mesh.face_node_connectivity.indices_by_location(),
        mesh.face_node_connectivity.start_index,
        elem_coords=elem_coords,
        location=location,
        mask=mask,
    )
    return meshinfo


def _regrid_along_dims(data, regridder, dims, num_out_dims, mdtol):
    """
    Perform regridding on data over specific dimensions.

    Parameters
    ----------
    data : array
        The data to be regrid.
    regridder : Regridder
        An instance of Regridder initialised to perform regridding.
    dims : tuple of int
        The dimensions in the source data to regrid along.
    num_out_dims : int
        The dimensionality of the target grid/mesh.
    mdtol : float
        Tolerance of missing data.

    Returns
    -------
    array
        The result of regridding the data.

    """
    num_dims = len(dims)
    standard_in_dims = [-1, -2][:num_dims]
    data = np.moveaxis(data, dims, standard_in_dims)
    result = regridder.regrid(data, mdtol=mdtol)

    standard_out_dims = [-1, -2][:num_out_dims]
    if num_dims == 2 and num_out_dims == 1:
        dims = [min(dims)]
    if num_dims == 1 and num_out_dims == 2:
        dims = [dims[0] + 1, dims[0]]
    result = np.moveaxis(result, standard_out_dims, dims)
    return result


def _map_complete_blocks(
    src, func, active_dims, out_sizes, *args, dtype=None, **kwargs
):
    """
    Apply a function to complete blocks.

    Based on :func:`iris._lazy_data.map_complete_blocks`.
    By "complete blocks" we mean that certain dimensions are enforced to be
    spanned by single chunks.
    Unlike the iris version of this function, this function also handles
    cases where the input and output have a different number of dimensions.
    The particular cases this function is designed for involves collapsing
    a 2D grid to a 1D mesh and expanding a 1D mesh to a 2D grid. Cases
    involving the mapping between the same number of dimensions should still
    behave the same as before.

    Parameters
    ----------
    src : cube
        Source :class:`~iris.cube.Cube` that function is applied to.
    func : function
        Function to apply.
    active_dims : tuple of int
        Dimensions that cannot be chunked.
    out_sizes : tuple of int
        Output size of dimensions that cannot be chunked.
    dtype : type, optional
        Type of the output array, if not given, the dtype of src is used.

    Returns
    -------
    array
        Either a :class:`dask.array.array`, or :class:`numpy.ndarray`
        depending on the laziness of the data in src.

    """
    if not src.has_lazy_data():
        return func(src.data, *args, **kwargs)

    data = src.lazy_data()
    if dtype is None:
        dtype = data.dtype

    # Ensure dims are not chunked
    in_chunks = list(data.chunks)
    for dim in active_dims:
        in_chunks[dim] = src.shape[dim]
    data = data.rechunk(in_chunks)

    # Determine output chunks
    num_dims = len(active_dims)
    num_out = len(out_sizes)
    out_chunks = list(data.chunks)
    sorted_dims = sorted(active_dims)
    if num_out == 1:
        out_chunks[sorted_dims[0]] = out_sizes[0]
    else:
        for dim, size in zip(active_dims, out_sizes):
            # Note: when mapping 2D to 2D, this will be the only alteration to
            # out_chunks, the same as iris._lazy_data.map_complete_blocks
            out_chunks[dim] = size

    dropped_dims = []
    new_axis = None
    if num_out > num_dims:
        # While this code should be robust for cases where num_out > num_dims > 1,
        # there is some ambiguity as to what their behaviour ought to be.
        # Since these cases are out of our own scope, we explicitly ignore them
        # for the time being.
        assert num_dims == 1
        # While this code should be robust for cases where num_out > 2,
        # we expect to handle at most 2D grids.
        # Since these cases are out of our own scope, we explicitly ignore them
        # for the time being.
        assert num_out == 2
        slice_index = sorted_dims[-1]
        # Insert the remaining contents of out_sizes in the position immediately
        # after the last dimension.
        out_chunks[slice_index:slice_index] = out_sizes[num_dims:]
        new_axis = range(slice_index, slice_index + num_out - num_dims)
    elif num_dims > num_out:
        # While this code should be robust for cases where num_dims > num_out > 1,
        # there is some ambiguity as to what their behaviour ought to be.
        # Since these cases are out of our own scope, we explicitly ignore them
        # for the time being.
        assert num_out == 1
        # While this code should be robust for cases where num_dims > 2,
        # we expect to handle at most 2D grids.
        # Since these cases are out of our own scope, we explicitly ignore them
        # for the time being.
        assert num_dims == 2
        dropped_dims = sorted_dims[num_out:]
        # Remove the remaining dimensions from the expected output shape.
        for dim in dropped_dims[::-1]:
            out_chunks.pop(dim)
    else:
        pass

    return data.map_blocks(
        func,
        *args,
        chunks=out_chunks,
        drop_axis=dropped_dims,
        new_axis=new_axis,
        dtype=dtype,
        **kwargs,
    )


def _create_cube(data, src_cube, src_dims, tgt_coords, num_tgt_dims):
    r"""
    Return a new cube for the result of regridding.

    Returned cube represents the result of regridding the source cube
    onto the new grid/mesh.
    All the metadata and coordinates of the result cube are copied from
    the source cube, with two exceptions:
        - Grid/mesh coordinates are copied from the target cube.
        - Auxiliary coordinates which span the grid dimensions are
          ignored.

    Parameters
    ----------
    data : array
        The regridded data as an N-dimensional NumPy array.
    src_cube : cube
        The source Cube.
    src_dims : tuple of int
        The dimensions of the X and Y coordinate within the source Cube.
    tgt_coords : tuple of :class:`iris.coords.Coord`\\ 's
        Either two 1D :class:`iris.coords.DimCoord`\\ 's, two 1D
        :class:`iris.experimental.ugrid.DimCoord`\\ 's or two 2D
        :class:`iris.coords.AuxCoord`\\ 's representing the new grid's
        X and Y coordinates.
    num_tgt_dims : int
        The number of dimensions that the target grid/mesh spans.

    Returns
    -------
    cube
        A new iris.cube.Cube instance.

    """
    new_cube = iris.cube.Cube(data)

    if len(src_dims) == 2:
        grid_x_dim, grid_y_dim = src_dims
    elif len(src_dims) == 1:
        grid_y_dim = src_dims[0]
        grid_x_dim = grid_y_dim + 1
    else:
        raise ValueError(
            f"Source grid must be described by 1 or 2 dimensions, got {len(src_dims)}"
        )
    if num_tgt_dims == 1:
        grid_y_dim = grid_x_dim = min(src_dims)
    for tgt_coord, dim in zip(tgt_coords, (grid_x_dim, grid_y_dim)):
        if len(tgt_coord.shape) == 1:
            if isinstance(tgt_coord, iris.coords.DimCoord):
                new_cube.add_dim_coord(tgt_coord, dim)
            else:
                new_cube.add_aux_coord(tgt_coord, dim)
        else:
            new_cube.add_aux_coord(tgt_coord, (grid_y_dim, grid_x_dim))

    new_cube.metadata = copy.deepcopy(src_cube.metadata)

    def copy_coords(src_coords, add_method):
        for coord in src_coords:
            dims = src_cube.coord_dims(coord)
            if set(src_dims).intersection(set(dims)):
                continue
            offset = num_tgt_dims - len(src_dims)
            dims = [dim if dim < max(src_dims) else dim + offset for dim in dims]
            result_coord = coord.copy()
            # Add result_coord to the owner of add_method.
            add_method(result_coord, dims)

    copy_coords(src_cube.dim_coords, new_cube.add_dim_coord)
    copy_coords(src_cube.aux_coords, new_cube.add_aux_coord)

    return new_cube


RegridInfo = namedtuple("RegridInfo", ["dims", "target", "regridder"])
GridRecord = namedtuple("GridRecord", ["grid_x", "grid_y"])
MeshRecord = namedtuple("MeshRecord", ["mesh", "location"])


def _make_gridinfo(cube, method, resolution, mask):
    method = check_method(method)
    if resolution is not None:
        if not (isinstance(resolution, int) and resolution > 0):
            raise ValueError("resolution must be a positive integer.")
        if method != Constants.Method.CONSERVATIVE:
            raise ValueError("resolution can only be set for conservative regridding.")
    if method == Constants.Method.CONSERVATIVE:
        center = False
    elif method in (Constants.Method.NEAREST, Constants.Method.BILINEAR):
        center = True
    return _cube_to_GridInfo(cube, center=center, resolution=resolution, mask=mask)


def _make_meshinfo(cube_or_mesh, method, mask, src_or_tgt, location=None):
    method = check_method(method)
    if isinstance(cube_or_mesh, Mesh):
        mesh = cube_or_mesh
    else:
        mesh = cube_or_mesh.mesh
        location = cube_or_mesh.location
    if mesh is None:
        raise ValueError(f"The {src_or_tgt} cube is not defined on a mesh.")
    if method == Constants.Method.CONSERVATIVE:
        if location != "face":
            raise ValueError(
                f"{method.name.lower()} regridding requires a {src_or_tgt} cube located on "
                f"the face of a cube, target cube had the {location} location."
            )
    elif method in (Constants.Method.NEAREST, Constants.Method.BILINEAR):
        if location not in ["face", "node"]:
            raise ValueError(
                f"{method.name.lower()} regridding requires a {src_or_tgt} cube with a node "
                f"or face location, target cube had the {location} location."
            )
        if location == "face" and None in mesh.face_coords:
            raise ValueError(
                f"{method.name.lower()} regridding requires a {src_or_tgt} cube on a face"
                f"location to have a face center."
            )

    return _mesh_to_MeshInfo(mesh, location, mask=mask)


def _regrid_rectilinear_to_rectilinear__prepare(
    src_grid_cube,
    tgt_grid_cube,
    method,
    precomputed_weights=None,
    src_resolution=None,
    tgt_resolution=None,
    src_mask=None,
    tgt_mask=None,
):
    """
    First (setup) part of 'regrid_rectilinear_to_rectilinear'.

    Check inputs and calculate the sparse regrid matrix and related info.
    The 'regrid info' returned can be re-used over many 2d slices.

    """
    tgt_x = _get_coord(tgt_grid_cube, "x")
    tgt_y = _get_coord(tgt_grid_cube, "y")
    src_x = _get_coord(src_grid_cube, "x")
    src_y = _get_coord(src_grid_cube, "y")

    if len(src_x.shape) == 1:
        grid_x_dim = src_grid_cube.coord_dims(src_x)[0]
        grid_y_dim = src_grid_cube.coord_dims(src_y)[0]
    else:
        grid_y_dim, grid_x_dim = src_grid_cube.coord_dims(src_x)

    srcinfo = _make_gridinfo(src_grid_cube, method, src_resolution, src_mask)
    tgtinfo = _make_gridinfo(tgt_grid_cube, method, tgt_resolution, tgt_mask)

    regridder = Regridder(
        srcinfo, tgtinfo, method=method, precomputed_weights=precomputed_weights
    )

    regrid_info = RegridInfo(
        dims=[grid_x_dim, grid_y_dim],
        target=GridRecord(tgt_x, tgt_y),
        regridder=regridder,
    )

    return regrid_info


def _regrid_rectilinear_to_rectilinear__perform(src_cube, regrid_info, mdtol):
    """
    Second (regrid) part of 'regrid_rectilinear_to_rectilinear'.

    Perform the prepared regrid calculation on a single cube.

    """
    grid_x_dim, grid_y_dim = regrid_info.dims
    grid_x, grid_y = regrid_info.target
    regridder = regrid_info.regridder

    out_dtype = regridder._out_dtype(src_cube.dtype)

    # Apply regrid to all the chunks of src_cube, ensuring first that all
    # chunks cover the entire horizontal plane (otherwise they would break
    # the regrid function).
    if len(grid_x.shape) == 1:
        chunk_shape = (len(grid_x.points), len(grid_y.points))
    else:
        # Due to structural reasons, the order here must be reversed.
        chunk_shape = grid_x.shape[::-1]
    new_data = _map_complete_blocks(
        src_cube,
        _regrid_along_dims,
        (grid_x_dim, grid_y_dim),
        chunk_shape,
        regridder=regridder,
        dims=[grid_x_dim, grid_y_dim],
        num_out_dims=2,
        mdtol=mdtol,
        dtype=out_dtype,
    )

    new_cube = _create_cube(
        new_data,
        src_cube,
        (grid_x_dim, grid_y_dim),
        (grid_x, grid_y),
        2,
    )
    return new_cube


def _regrid_unstructured_to_rectilinear__prepare(
    src_mesh_cube,
    target_grid_cube,
    method,
    precomputed_weights=None,
    tgt_resolution=None,
    src_mask=None,
    tgt_mask=None,
):
    """
    First (setup) part of 'regrid_unstructured_to_rectilinear'.

    Check inputs and calculate the sparse regrid matrix and related info.
    The 'regrid info' returned can be re-used over many 2d slices.

    """
    grid_x = _get_coord(target_grid_cube, "x")
    grid_y = _get_coord(target_grid_cube, "y")

    # From src_mesh_cube, fetch the mesh, and the dimension on the cube which that
    # mesh belongs to.
    mesh_dim = src_mesh_cube.mesh_dim()

    meshinfo = _make_meshinfo(src_mesh_cube, method, src_mask, "source")
    gridinfo = _make_gridinfo(target_grid_cube, method, tgt_resolution, tgt_mask)

    regridder = Regridder(
        meshinfo, gridinfo, method=method, precomputed_weights=precomputed_weights
    )

    regrid_info = RegridInfo(
        dims=[mesh_dim],
        target=GridRecord(grid_x, grid_y),
        regridder=regridder,
    )

    return regrid_info


def _regrid_unstructured_to_rectilinear__perform(src_cube, regrid_info, mdtol):
    """
    Second (regrid) part of 'regrid_unstructured_to_rectilinear'.

    Perform the prepared regrid calculation on a single cube.

    """
    (mesh_dim,) = regrid_info.dims
    grid_x, grid_y = regrid_info.target
    regridder = regrid_info.regridder

    out_dtype = regridder._out_dtype(src_cube.dtype)

    # Apply regrid to all the chunks of src_cube, ensuring first that all
    # chunks cover the entire horizontal plane (otherwise they would break
    # the regrid function).
    if len(grid_x.shape) == 1:
        chunk_shape = (len(grid_x.points), len(grid_y.points))
    else:
        # Due to structural reasons, the order here must be reversed.
        chunk_shape = grid_x.shape[::-1]
    new_data = _map_complete_blocks(
        src_cube,
        _regrid_along_dims,
        (mesh_dim,),
        chunk_shape,
        regridder=regridder,
        dims=[mesh_dim],
        num_out_dims=2,
        mdtol=mdtol,
        dtype=out_dtype,
    )

    new_cube = _create_cube(
        new_data,
        src_cube,
        (mesh_dim,),
        (grid_x, grid_y),
        2,
    )

    # TODO: apply tweaks to created cube (slice out length 1 dimensions)

    return new_cube


def _regrid_rectilinear_to_unstructured__prepare(
    src_grid_cube,
    tgt_cube_or_mesh,
    method,
    precomputed_weights=None,
    src_resolution=None,
    src_mask=None,
    tgt_mask=None,
    tgt_location=None,
):
    """
    First (setup) part of 'regrid_rectilinear_to_unstructured'.

    Check inputs and calculate the sparse regrid matrix and related info.
    The 'regrid info' returned can be re-used over many 2d slices.

    """
    grid_x = _get_coord(src_grid_cube, "x")
    grid_y = _get_coord(src_grid_cube, "y")
    if isinstance(tgt_cube_or_mesh, Mesh):
        mesh = tgt_cube_or_mesh
        location = tgt_location
    else:
        mesh = tgt_cube_or_mesh.mesh
        location = tgt_cube_or_mesh.location

    if grid_x.ndim == 1:
        (grid_x_dim,) = src_grid_cube.coord_dims(grid_x)
        (grid_y_dim,) = src_grid_cube.coord_dims(grid_y)
    else:
        grid_y_dim, grid_x_dim = src_grid_cube.coord_dims(grid_x)

    meshinfo = _make_meshinfo(
        tgt_cube_or_mesh, method, tgt_mask, "target", location=tgt_location
    )
    gridinfo = _make_gridinfo(src_grid_cube, method, src_resolution, src_mask)

    regridder = Regridder(
        gridinfo, meshinfo, method=method, precomputed_weights=precomputed_weights
    )

    regrid_info = RegridInfo(
        dims=[grid_x_dim, grid_y_dim],
        target=MeshRecord(mesh, location),
        regridder=regridder,
    )

    return regrid_info


def _regrid_rectilinear_to_unstructured__perform(src_cube, regrid_info, mdtol):
    """
    Second (regrid) part of 'regrid_rectilinear_to_unstructured'.

    Perform the prepared regrid calculation on a single cube.

    """
    grid_x_dim, grid_y_dim = regrid_info.dims
    mesh, location = regrid_info.target
    regridder = regrid_info.regridder

    if location == "face":
        face_node = mesh.face_node_connectivity
        # In face_node_connectivity: `location`= face, `connected` = node, so
        # you want to get the length of the `location` dimension.
        chunk_shape = (face_node.shape[face_node.location_axis],)
    elif location == "node":
        chunk_shape = mesh.node_coords[0].shape
    else:
        raise NotImplementedError(f"Unrecognised location {location}.")

    out_dtype = regridder._out_dtype(src_cube.dtype)

    # Apply regrid to all the chunks of src_cube, ensuring first that all
    # chunks cover the entire horizontal plane (otherwise they would break
    # the regrid function).
    new_data = _map_complete_blocks(
        src_cube,
        _regrid_along_dims,
        (grid_x_dim, grid_y_dim),
        chunk_shape,
        regridder=regridder,
        dims=[grid_x_dim, grid_y_dim],
        num_out_dims=1,
        mdtol=mdtol,
        dtype=out_dtype,
    )

    new_cube = _create_cube(
        new_data,
        src_cube,
        (grid_x_dim, grid_y_dim),
        mesh.to_MeshCoords(location),
        1,
    )
    return new_cube


def _regrid_unstructured_to_unstructured__prepare(
    src_mesh_cube,
    tgt_cube_or_mesh,
    method,
    precomputed_weights=None,
    src_mask=None,
    tgt_mask=None,
    src_location=None,
    tgt_location=None,
):
    """
    First (setup) part of 'regrid_unstructured_to_unstructured'.

    Check inputs and calculate the sparse regrid matrix and related info.
    The 'regrid info' returned can be re-used over many 2d slices.

    """
    if isinstance(tgt_cube_or_mesh, Mesh):
        mesh = tgt_cube_or_mesh
        location = tgt_location
    else:
        mesh = tgt_cube_or_mesh.mesh
        location = tgt_cube_or_mesh.location

    mesh_dim = src_mesh_cube.mesh_dim()

    src_meshinfo = _make_meshinfo(
        src_mesh_cube, method, src_mask, "source", location=src_location
    )
    tgt_meshinfo = _make_meshinfo(
        tgt_cube_or_mesh, method, tgt_mask, "target", location=tgt_location
    )

    regridder = Regridder(
        src_meshinfo,
        tgt_meshinfo,
        method=method,
        precomputed_weights=precomputed_weights,
    )

    regrid_info = RegridInfo(
        dims=[mesh_dim],
        target=MeshRecord(mesh, location),
        regridder=regridder,
    )

    return regrid_info


def _regrid_unstructured_to_unstructured__perform(src_cube, regrid_info, mdtol):
    """
    Second (regrid) part of 'regrid_unstructured_to_unstructured'.

    Perform the prepared regrid calculation on a single cube.

    """
    (mesh_dim,) = regrid_info.dims
    mesh, location = regrid_info.target
    regridder = regrid_info.regridder

    out_dtype = regridder._out_dtype(src_cube.dtype)

    if location == "face":
        face_node = mesh.face_node_connectivity
        chunk_shape = (face_node.shape[face_node.location_axis],)
    elif location == "node":
        chunk_shape = mesh.node_coords[0].shape
    else:
        raise NotImplementedError(f"Unrecognised location {location}.")

    new_data = _map_complete_blocks(
        src_cube,
        _regrid_along_dims,
        (mesh_dim,),
        chunk_shape,
        regridder=regridder,
        dims=[mesh_dim],
        num_out_dims=1,
        mdtol=mdtol,
        dtype=out_dtype,
    )

    new_cube = _create_cube(
        new_data,
        src_cube,
        (mesh_dim,),
        mesh.to_MeshCoords(location),
        1,
    )

    return new_cube


def regrid_rectilinear_to_rectilinear(
    src_cube,
    grid_cube,
    mdtol=0,
    method=Constants.Method.CONSERVATIVE,
    src_resolution=None,
    tgt_resolution=None,
):
    r"""
    Regrid rectilinear :class:`~iris.cube.Cube` onto another rectilinear grid.

    Return a new :class:`~iris.cube.Cube` with :attr:`~iris.cube.Cube.data`
    values calculated using the area weighted
    mean of :attr:`~iris.cube.Cube.data` values from ``src_cube`` regridded onto the
    horizontal grid of ``grid_cube``.

    Parameters
    ----------
    src_cube : :class:`iris.cube.Cube`
        A rectilinear instance of :class:`~iris.cube.Cube` that supplies the data,
        metadata and coordinates.
    grid_cube : :class:`iris.cube.Cube`
        A rectilinear instance of :class:`~iris.cube.Cube` that supplies the desired
        horizontal grid definition.
    mdtol : float, default=0
        Tolerance of missing data. The value returned in each element of the
        returned :class:`~iris.cube.Cube`\\ 's :attr:`~iris.cube.Cube.data`
        array will be masked if the fraction of masked
        data in the overlapping cells of ``src_cube`` exceeds ``mdtol``. This
        fraction is calculated based on the area of masked cells within each
        target cell. ``mdtol=0`` means no missing data is tolerated while ``mdtol=1``
        will mean the resulting element will be masked if and only if all the
        overlapping cells of ``src_cube`` are masked.
    method : :class:`Constants.Method`, , default=Constants.Method.CONSERVATIVE.
            The method to be used to calculate weights.
    src_resolution, tgt_resolution : int, optional
        If present, represents the amount of latitude slices per source/target cell
        given to ESMF for calculation.

    Returns
    -------
    :class:`iris.cube.Cube`
        A new :class:`~iris.cube.Cube` instance.

    """
    regrid_info = _regrid_rectilinear_to_rectilinear__prepare(
        src_cube,
        grid_cube,
        method=method,
        src_resolution=src_resolution,
        tgt_resolution=tgt_resolution,
    )
    result = _regrid_rectilinear_to_rectilinear__perform(src_cube, regrid_info, mdtol)
    return result


class ESMFAreaWeighted:
    """
    A scheme which can be recognised by :meth:`iris.cube.Cube.regrid`.

    This class describes an area-weighted regridding scheme for regridding
    between horizontal grids/meshes. It uses :mod:`esmpy` to handle
    calculations and allows for different coordinate systems.

    """

    def __init__(
        self, mdtol=0, use_src_mask=False, use_tgt_mask=False, tgt_location="face"
    ):
        """
        Area-weighted scheme for regridding between rectilinear grids.

        Parameters
        ----------
        mdtol : float, default=0
            Tolerance of missing data. The value returned in each element of
            the returned array will be masked if the fraction of missing data
            exceeds ``mdtol``. This fraction is calculated based on the area of
            masked cells within each target cell. ``mdtol=0`` means no masked
            data is tolerated while ``mdtol=1`` will mean the resulting element
            will be masked if and only if all the overlapping elements of the
            source grid are masked.
        use_src_mask : bool, default=False
            If True, derive a mask from source cube which will tell :mod:`esmpy`
            which points to ignore.
        use_tgt_mask : bool, default=False
            If True, derive a mask from target cube which will tell :mod:`esmpy`
            which points to ignore.
        tgt_location : str or None, default="face"
            Either "face" or "node". Describes the location for data on the mesh
            if the target is not a :class:`~iris.cube.Cube`.

        """
        if not (0 <= mdtol <= 1):
            msg = "Value for mdtol must be in range 0 - 1, got {}."
            raise ValueError(msg.format(mdtol))
        if tgt_location is not None and tgt_location != "face":
            raise ValueError(
                "For area weighted regridding, target location must be 'face'."
            )
        self.mdtol = mdtol
        self.use_src_mask = use_src_mask
        self.use_tgt_mask = use_tgt_mask
        self.tgt_location = "face"

    def __repr__(self):
        """Return a representation of the class."""
        return "ESMFAreaWeighted(mdtol={})".format(self.mdtol)

    def regridder(
        self,
        src_grid,
        tgt_grid,
        src_resolution=None,
        tgt_resolution=None,
        use_src_mask=None,
        use_tgt_mask=None,
        tgt_location="face",
    ):
        """
        Create regridder to perform regridding from ``src_grid`` to ``tgt_grid``.

        Parameters
        ----------
        src_grid : :class:`iris.cube.Cube`
            The :class:`~iris.cube.Cube` defining the source.
        tgt_grid : :class:`iris.cube.Cube` or :class:`iris.experimental.ugrid.Mesh`
            The unstructured :class:`~iris.cube.Cube`or
            :class:`~iris.experimental.ugrid.Mesh` defining the target.
        src_resolution, tgt_resolution : int, optional
            If present, represents the amount of latitude slices per source/target cell
            given to ESMF for calculation. If resolution is set, ``src`` and ``tgt``
            respectively must have strictly increasing bounds (bounds may be transposed
            plus or minus 360 degrees to make the bounds strictly increasing).
        use_src_mask : :obj:`~numpy.typing.ArrayLike` or bool, optional
            Array describing which elements :mod:`esmpy` will ignore on the src_grid.
            If True, the mask will be derived from src_grid.
        use_tgt_mask : :obj:`~numpy.typing.ArrayLike` or bool, optional
            Array describing which elements :mod:`esmpy` will ignore on the tgt_grid.
            If True, the mask will be derived from tgt_grid.
        tgt_location : str or None, default="face"
            Either "face" or "node". Describes the location for data on the mesh
            if the target is not a :class:`~iris.cube.Cube`.


        Returns
        -------
        :class:`ESMFAreaWeightedRegridder`
            A callable instance of a regridder with the interface: ``regridder(cube)``
                ... where ``cube`` is a :class:`~iris.cube.Cube` with the same
                grid as ``src_grid`` that is to be regridded to the grid of
                ``tgt_grid``.

        Raises
        ------
        ValueError
            If ``use_src_mask`` or ``use_tgt_mask`` are True while the masks on ``src_grid``
            or ``tgt_grid`` respectively are not constant over non-horizontal dimensions.
        """
        if use_src_mask is None:
            use_src_mask = self.use_src_mask
        if use_tgt_mask is None:
            use_tgt_mask = self.use_tgt_mask
        if tgt_location is not None and tgt_location != "face":
            raise ValueError(
                "For area weighted regridding, target location must be 'face'."
            )
        return ESMFAreaWeightedRegridder(
            src_grid,
            tgt_grid,
            mdtol=self.mdtol,
            src_resolution=src_resolution,
            tgt_resolution=tgt_resolution,
            use_src_mask=use_src_mask,
            use_tgt_mask=use_tgt_mask,
            tgt_location="face",
        )


class ESMFBilinear:
    """
    A scheme which can be recognised by :meth:`iris.cube.Cube.regrid`.

    This class describes a bilinear regridding scheme for regridding
    between horizontal grids/meshes. It uses :mod:`esmpy` to handle
    calculations and allows for different coordinate systems.
    """

    def __init__(
        self, mdtol=0, use_src_mask=False, use_tgt_mask=False, tgt_location=None
    ):
        """
        Area-weighted scheme for regridding between rectilinear grids.

        Parameters
        ----------
        mdtol : float, default=0
            Tolerance of missing data. The value returned in each element of
            the returned array will be masked if the fraction of missing data
            exceeds ``mdtol``.
        use_src_mask : bool, default=False
            If True, derive a mask from source cube which will tell :mod:`esmpy`
            which points to ignore.
        use_tgt_mask : bool, default=False
            If True, derive a mask from target cube which will tell :mod:`esmpy`
            which points to ignore.
        tgt_location : str or None, default=None
            Either "face" or "node". Describes the location for data on the mesh
            if the target is not a :class:`~iris.cube.Cube`.

        """
        if not (0 <= mdtol <= 1):
            msg = "Value for mdtol must be in range 0 - 1, got {}."
            raise ValueError(msg.format(mdtol))
        self.mdtol = mdtol
        self.use_src_mask = use_src_mask
        self.use_tgt_mask = use_tgt_mask
        self.tgt_location = tgt_location

    def __repr__(self):
        """Return a representation of the class."""
        return "ESMFBilinear(mdtol={})".format(self.mdtol)

    def regridder(
        self,
        src_grid,
        tgt_grid,
        use_src_mask=None,
        use_tgt_mask=None,
        tgt_location=None,
    ):
        """
        Create regridder to perform regridding from ``src_grid`` to ``tgt_grid``.

        Parameters
        ----------
        src_grid : :class:`iris.cube.Cube`
            The :class:`~iris.cube.Cube` defining the source.
        tgt_grid : :class:`iris.cube.Cube` or :class:`iris.experimental.ugrid.Mesh`
            The unstructured :class:`~iris.cube.Cube`or
            :class:`~iris.experimental.ugrid.Mesh` defining the target.
        use_src_mask : :obj:`~numpy.typing.ArrayLike` or bool, optional
            Array describing which elements :mod:`esmpy` will ignore on the src_grid.
            If True, the mask will be derived from src_grid.
        use_tgt_mask : :obj:`~numpy.typing.ArrayLike` or bool, optional
            Array describing which elements :mod:`esmpy` will ignore on the tgt_grid.
            If True, the mask will be derived from tgt_grid.
        tgt_location : str or None, default=None
            Either "face" or "node". Describes the location for data on the mesh
            if the target is not a :class:`~iris.cube.Cube`.

        Returns
        -------
        :class:`ESMFBilinearRegridder`
            A callable instance of a regridder with the interface: ``regridder(cube)``
                ... where ``cube`` is a :class:`~iris.cube.Cube` with the same
                grid as ``src_grid`` that is to be regridded to the grid of
                ``tgt_grid``.

        Raises
        ------
        ValueError
            If ``use_src_mask`` or ``use_tgt_mask`` are True while the masks on ``src_grid``
            or ``tgt_grid`` respectively are not constant over non-horizontal dimensions.
        """
        if use_src_mask is None:
            use_src_mask = self.use_src_mask
        if use_tgt_mask is None:
            use_tgt_mask = self.use_tgt_mask
        if tgt_location is None:
            tgt_location = self.tgt_location
        return ESMFBilinearRegridder(
            src_grid,
            tgt_grid,
            mdtol=self.mdtol,
            use_src_mask=use_src_mask,
            use_tgt_mask=use_tgt_mask,
            tgt_location=tgt_location,
        )


class ESMFNearest:
    """
    A scheme which can be recognised by :meth:`iris.cube.Cube.regrid`.

    This class describes a nearest neighbour regridding scheme for regridding
    between horizontal grids/meshes. It uses :mod:`esmpy` to handle
    calculations and allows for different coordinate systems.

    Notes
    -----
    By default, masked data is handled by masking the result when the
    nearest source point is masked, however, if `use_src_mask` is True, then
    the nearest unmasked point is used instead. This differs from other regridding
    methods where `use_src_mask` has no effect on the mathematics. Like other
    methods, when `use_src_mask` is True, the mask must be constant over all
    non-horizontal dimensions otherwise an error is thrown.

    When two source points are a constant distance from a target point,
    the decision for which point is used is based on the indexing which ESMF gives
    the points. ESMF describes this decision as somewhat arbitrary:
    https://earthsystemmodeling.org/docs/release/latest/ESMF_refdoc/node3.html#SECTION03023000000000000000
    As a result of this, while the same object ought to behave consistently for
    this scheme, there is no guarantee that different objects representing
    the same equivalent space will behave the same.
    """

    def __init__(self, use_src_mask=False, use_tgt_mask=False, tgt_location=None):
        """
        Nearest neighbour scheme for regridding between rectilinear grids.

        Parameters
        ----------
        use_src_mask : bool, default=False
            If True, derive a mask from source cube which will tell :mod:`esmpy`
            which points to ignore.
        use_tgt_mask : bool, default=False
            If True, derive a mask from target cube which will tell :mod:`esmpy`
            which points to ignore.
        tgt_location : str or None, default=None
            Either "face" or "node". Describes the location for data on the mesh
            if the target is not a :class:`~iris.cube.Cube`.
        """
        self.use_src_mask = use_src_mask
        self.use_tgt_mask = use_tgt_mask
        self.tgt_location = tgt_location

    def __repr__(self):
        """Return a representation of the class."""
        return "ESMFNearest()"

    def regridder(
        self,
        src_grid,
        tgt_grid,
        use_src_mask=None,
        use_tgt_mask=None,
        tgt_location=None,
    ):
        """
        Create regridder to perform regridding from ``src_grid`` to ``tgt_grid``.

        Parameters
        ----------
        src_grid : :class:`iris.cube.Cube`
            The :class:`~iris.cube.Cube` defining the source.
        tgt_grid : :class:`iris.cube.Cube` or :class:`iris.experimental.ugrid.Mesh`
            The unstructured :class:`~iris.cube.Cube`or
            :class:`~iris.experimental.ugrid.Mesh` defining the target.
        use_src_mask : :obj:`~numpy.typing.ArrayLike` or bool, optional
            Array describing which elements :mod:`esmpy` will ignore on the src_grid.
            If True, the mask will be derived from src_grid.
        use_tgt_mask : :obj:`~numpy.typing.ArrayLike` or bool, optional
            Array describing which elements :mod:`esmpy` will ignore on the tgt_grid.
            If True, the mask will be derived from tgt_grid.
        tgt_location : str or None, default=None
            Either "face" or "node". Describes the location for data on the mesh
            if the target is not a :class:`~iris.cube.Cube`.

        Returns
        -------
        :class:`ESMFNearestRegridder`
            A callable instance of a regridder with the interface: ``regridder(cube)``
                ... where ``cube`` is a :class:`~iris.cube.Cube` with the same
                grid as ``src_grid`` that is to be regridded to the grid of
                ``tgt_grid``.

        Raises
        ------
        ValueError
            If ``use_src_mask`` or ``use_tgt_mask`` are True while the masks on ``src_grid``
            or ``tgt_grid`` respectively are not constant over non-horizontal dimensions.
        """
        if use_src_mask is None:
            use_src_mask = self.use_src_mask
        if use_tgt_mask is None:
            use_tgt_mask = self.use_tgt_mask
        if tgt_location is None:
            tgt_location = self.tgt_location
        return ESMFNearestRegridder(
            src_grid,
            tgt_grid,
            use_src_mask=use_src_mask,
            use_tgt_mask=use_tgt_mask,
            tgt_location=tgt_location,
        )


class _ESMFRegridder:
    r"""Generic regridder class for regridding of :class:`~iris.cube.Cube`\\ s."""

    def __init__(
        self,
        src,
        tgt,
        method,
        mdtol=None,
        use_src_mask=False,
        use_tgt_mask=False,
        tgt_location=None,
        **kwargs,
    ):
        """
        Create regridder for conversions between ``src`` and ``tgt``.

        Parameters
        ----------
        src : :class:`iris.cube.Cube`
            The rectilinear :class:`~iris.cube.Cube` providing the source grid.
        tgt : :class:`iris.cube.Cube` or :class:`iris.experimental.ugrid.Mesh`
            The rectilinear :class:`~iris.cube.Cube` providing the target grid.
        method : :class:`Constants.Method`
            The method to be used to calculate weights.
        mdtol : float, default=None
            Tolerance of missing data. The value returned in each element of
            the returned array will be masked if the fraction of masked data
            exceeds ``mdtol``. ``mdtol=0`` means no missing data is tolerated while
            ``mdtol=1`` will mean the resulting element will be masked if and only
            if all the contributing elements of data are masked. If no value is given,
            this will default to 1 for conservative regridding and 0 otherwise.
        use_src_mask, use_tgt_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
            Either an array representing the cells in the source/target to ignore, or else
            a boolean value. If True, this array is taken from the mask on the data
            in ``src`` or ``tgt``. If False, no mask will be taken and all points will
            be used in weights calculation.
        tgt_location : str or None, default=None
            Either "face" or "node". Describes the location for data on the mesh
            if the target is not a :class:`~iris.cube.Cube`.

        Raises
        ------
        ValueError
            If ``use_src_mask`` or ``use_tgt_mask`` are True while the masks on ``src``
            or ``tgt`` respectively are not constant over non-horizontal dimensions.

        """
        method = check_method(method)
        if mdtol is None:
            if method == Constants.Method.CONSERVATIVE:
                mdtol = 1
            elif method in (Constants.Method.BILINEAR, Constants.Method.NEAREST):
                mdtol = 0
        if not (0 <= mdtol <= 1):
            msg = "Value for mdtol must be in range 0 - 1, got {}."
            raise ValueError(msg.format(mdtol))
        self.mdtol = mdtol
        self.method = method

        self.src_mask = _get_mask(src, use_src_mask)
        kwargs["src_mask"] = self.src_mask
        self.tgt_mask = _get_mask(tgt, use_tgt_mask)
        kwargs["tgt_mask"] = self.tgt_mask

        src_is_mesh = src.mesh is not None
        tgt_is_mesh = isinstance(tgt, Mesh) or tgt.mesh is not None
        if src_is_mesh:
            if tgt_is_mesh:
                prepare_func = _regrid_unstructured_to_unstructured__prepare
                kwargs["tgt_location"] = tgt_location
            else:
                prepare_func = _regrid_unstructured_to_rectilinear__prepare
        else:
            if tgt_is_mesh:
                prepare_func = _regrid_rectilinear_to_unstructured__prepare
                kwargs["tgt_location"] = tgt_location
            else:
                prepare_func = _regrid_rectilinear_to_rectilinear__prepare
        regrid_info = prepare_func(src, tgt, method, **kwargs)

        # Store regrid info.
        self._tgt = regrid_info.target
        self.regridder = regrid_info.regridder

        # Record the source grid.
        if src_is_mesh:
            self._src = MeshRecord(src.mesh, src.location)
        else:
            self._src = GridRecord(_get_coord(src, "x"), _get_coord(src, "y"))

    def __call__(self, cube):
        """
        Regrid this :class:`~iris.cube.Cube` onto the target grid of this regridder instance.

        The given :class:`~iris.cube.Cube` must be defined with the same grid as the source
        :class:`~iris.cube.Cube` used to create this :class:`_ESMFRegridder` instance.

        Parameters
        ----------
        cube : :class:`iris.cube.Cube`
            A :class:`~iris.cube.Cube` instance to be regridded.

        Returns
        -------
        :class:`iris.cube.Cube`
            A :class:`~iris.cube.Cube` defined with the horizontal dimensions of the target
            and the other dimensions from this :class:`~iris.cube.Cube`. The data values of
            this :class:`~iris.cube.Cube` will be converted to values on the new grid using
            regridding via :mod:`esmpy` generated weights.

        """
        if cube.mesh is not None:
            # TODO: replace temporary hack when iris issues are sorted.
            # Ignore differences in var_name that might be caused by saving.
            # TODO: uncomment this when iris issue with masked array comparison is sorted.
            # self_mesh = copy.deepcopy(self._src.mesh)
            # self_mesh.var_name = mesh.var_name
            # for self_coord, other_coord in zip(self_mesh.all_coords, mesh.all_coords):
            #     if self_coord is not None:
            #         self_coord.var_name = other_coord.var_name
            # for self_con, other_con in zip(
            #     self_mesh.all_connectivities, mesh.all_connectivities
            # ):
            #     if self_con is not None:
            #         self_con.var_name = other_con.var_name
            # if self_mesh != mesh:
            #     raise ValueError(
            #         "The given cube is not defined on the same "
            #         "source mesh as this regridder."
            #     )
            dims = [cube.mesh_dim()]

        else:
            new_src_x, new_src_y = (_get_coord(cube, "x"), _get_coord(cube, "y"))

            # Check the source grid matches that used in initialisation
            for saved_coord, new_coord in zip(self._src, (new_src_x, new_src_y)):
                # Ignore differences in var_name that might be caused by saving.
                saved_coord = copy.deepcopy(saved_coord)
                saved_coord.var_name = new_coord.var_name
                if saved_coord != new_coord:
                    raise ValueError(
                        "The given cube is not defined on the same "
                        "source grid as this regridder."
                    )

            if len(new_src_x.shape) == 1:
                dims = [cube.coord_dims(new_src_x)[0], cube.coord_dims(new_src_y)[0]]
            else:
                # Due to structural reasons, the order here must be reversed.
                dims = cube.coord_dims(new_src_x)[::-1]

        regrid_info = RegridInfo(
            dims=dims,
            target=self._tgt,
            regridder=self.regridder,
        )
        src_is_mesh = cube.mesh is not None
        tgt_is_mesh = isinstance(self._tgt, MeshRecord)

        if src_is_mesh:
            if tgt_is_mesh:
                perform_func = _regrid_unstructured_to_unstructured__perform
            else:
                perform_func = _regrid_unstructured_to_rectilinear__perform
        else:
            if tgt_is_mesh:
                perform_func = _regrid_rectilinear_to_unstructured__perform
            else:
                perform_func = _regrid_rectilinear_to_rectilinear__perform
        result = perform_func(cube, regrid_info, self.mdtol)

        return result


class ESMFAreaWeightedRegridder(_ESMFRegridder):
    r"""Regridder class for conservative regridding of :class:`~iris.cube.Cube`\\ s."""

    def __init__(
        self,
        src,
        tgt,
        mdtol=0,
        precomputed_weights=None,
        src_resolution=None,
        tgt_resolution=None,
        use_src_mask=False,
        use_tgt_mask=False,
        tgt_location="face",
    ):
        """
        Create regridder for conversions between ``src`` and ``tgt``.

        Parameters
        ----------
        src : :class:`iris.cube.Cube`
            The rectilinear :class:`~iris.cube.Cube` providing the source.
        tgt : :class:`iris.cube.Cube` or :class:`iris.experimental.ugrid.Mesh`
            The unstructured :class:`~iris.cube.Cube`or
            :class:`~iris.experimental.ugrid.Mesh` defining the target.
        mdtol : float, default=0
            Tolerance of missing data. The value returned in each element of
            the returned array will be masked if the fraction of masked data
            exceeds ``mdtol``. ``mdtol=0`` means no missing data is tolerated while
            ``mdtol=1`` will mean the resulting element will be masked if and only
            if all the contributing elements of data are masked.
        precomputed_weights : :class:`scipy.sparse.spmatrix`, optional
            If ``None``, :mod:`esmpy` will be used to
            calculate regridding weights. Otherwise, :mod:`esmpy` will be bypassed
            and ``precomputed_weights`` will be used as the regridding weights.
        src_resolution, tgt_resolution : int, optional
            If present, represents the amount of latitude slices per source/target cell
            given to ESMF for calculation. If resolution is set, ``src`` and ``tgt``
            respectively must have strictly increasing bounds (bounds may be transposed
            plus or minus 360 degrees to make the bounds strictly increasing).
        use_src_mask, use_tgt_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
            Either an array representing the cells in the source/target to ignore, or else
            a boolean value. If True, this array is taken from the mask on the data
            in ``src`` or ``tgt``. If False, no mask will be taken and all points will
            be used in weights calculation.
        tgt_location : str or None, default="face"
            Either "face" or "node". Describes the location for data on the mesh
            if the target is not a :class:`~iris.cube.Cube`.

        Raises
        ------
        ValueError
            If ``use_src_mask`` or ``use_tgt_mask`` are True while the masks on ``src``
            or ``tgt`` respectively are not constant over non-horizontal dimensions or
            if tgt_location is not "face".
        """
        kwargs = dict()
        self.src_resolution = src_resolution
        if src_resolution is not None:
            kwargs["src_resolution"] = src_resolution
        self.tgt_resolution = tgt_resolution
        if tgt_resolution is not None:
            kwargs["tgt_resolution"] = tgt_resolution
        if tgt_location is not None and tgt_location != "face":
            raise ValueError(
                "For area weighted regridding, target location must be 'face'."
            )
        kwargs["use_src_mask"] = use_src_mask
        kwargs["use_tgt_mask"] = use_tgt_mask
        super().__init__(
            src,
            tgt,
            Constants.Method.CONSERVATIVE,
            mdtol=mdtol,
            precomputed_weights=precomputed_weights,
            tgt_location="face",
            **kwargs,
        )


class ESMFBilinearRegridder(_ESMFRegridder):
    r"""Regridder class for bilinear regridding of :class:`~iris.cube.Cube`\\ s."""

    def __init__(
        self,
        src,
        tgt,
        mdtol=0,
        precomputed_weights=None,
        use_src_mask=False,
        use_tgt_mask=False,
        tgt_location=None,
    ):
        """
        Create regridder for conversions between ``src`` and ``tgt``.

        Parameters
        ----------
        src : :class:`iris.cube.Cube`
            The rectilinear :class:`~iris.cube.Cube` providing the source.
        tgt : :class:`iris.cube.Cube` or :class:`iris.experimental.ugrid.Mesh`
            The unstructured :class:`~iris.cube.Cube`or
            :class:`~iris.experimental.ugrid.Mesh` defining the target.
        mdtol : float, default=0
            Tolerance of missing data. The value returned in each element of
            the returned array will be masked if the fraction of masked data
            exceeds ``mdtol``. ``mdtol=0`` means no missing data is tolerated while
            ``mdtol=1`` will mean the resulting element will be masked if and only
            if all the contributing elements of data are masked.
        precomputed_weights : :class:`scipy.sparse.spmatrix`, optional
            If ``None``, :mod:`esmpy` will be used to
            calculate regridding weights. Otherwise, :mod:`esmpy` will be bypassed
            and ``precomputed_weights`` will be used as the regridding weights.
        use_src_mask, use_tgt_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
            Either an array representing the cells in the source/target to ignore, or else
            a boolean value. If True, this array is taken from the mask on the data
            in ``src`` or ``tgt``. If False, no mask will be taken and all points will
            be used in weights calculation.
        tgt_location : str or None, default=None
            Either "face" or "node". Describes the location for data on the mesh
            if the target is not a :class:`~iris.cube.Cube`.

        Raises
        ------
        ValueError
            If ``use_src_mask`` or ``use_tgt_mask`` are True while the masks on ``src``
            or ``tgt`` respectively are not constant over non-horizontal dimensions.
        """
        super().__init__(
            src,
            tgt,
            Constants.Method.BILINEAR,
            mdtol=mdtol,
            precomputed_weights=precomputed_weights,
            use_src_mask=use_src_mask,
            use_tgt_mask=use_tgt_mask,
            tgt_location=tgt_location,
        )


class ESMFNearestRegridder(_ESMFRegridder):
    r"""Regridder class for nearest (source to destination) regridding of :class:`~iris.cube.Cube`\\ s."""

    def __init__(
        self,
        src,
        tgt,
        precomputed_weights=None,
        use_src_mask=False,
        use_tgt_mask=False,
        tgt_location=None,
    ):
        """
        Create regridder for conversions between ``src`` and ``tgt``.

        Parameters
        ----------
        src : :class:`iris.cube.Cube`
            The rectilinear :class:`~iris.cube.Cube` providing the source.
        tgt : :class:`iris.cube.Cube` or :class:`iris.experimental.ugrid.Mesh`
            The unstructured :class:`~iris.cube.Cube`or
            :class:`~iris.experimental.ugrid.Mesh` defining the target.
        precomputed_weights : :class:`scipy.sparse.spmatrix`, optional
            If ``None``, :mod:`esmpy` will be used to
            calculate regridding weights. Otherwise, :mod:`esmpy` will be bypassed
            and ``precomputed_weights`` will be used as the regridding weights.
        use_src_mask, use_tgt_mask : :obj:`~numpy.typing.ArrayLike` or bool, default=False
            Either an array representing the cells in the source/target to ignore, or else
            a boolean value. If True, this array is taken from the mask on the data
            in ``src`` or ``tgt``. If False, no mask will be taken and all points will
            be used in weights calculation.
        tgt_location : str or None, default=None
            Either "face" or "node". Describes the location for data on the mesh
            if the target is not a :class:`~iris.cube.Cube`.

        Raises
        ------
        ValueError
            If ``use_src_mask`` or ``use_tgt_mask`` are True while the masks on ``src``
            or ``tgt`` respectively are not constant over non-horizontal dimensions.
        """
        super().__init__(
            src,
            tgt,
            Constants.Method.NEAREST,
            mdtol=0,
            precomputed_weights=precomputed_weights,
            use_src_mask=use_src_mask,
            use_tgt_mask=use_tgt_mask,
            tgt_location=tgt_location,
        )



"""Unit tests for :class:`esmf_regrid._esmf_sdo.RefinedGridInfo`."""

import numpy as np
from numpy import ma

from esmf_regrid.esmf_regridder import RefinedGridInfo, Regridder
from esmf_regrid.experimental.unstructured_regrid import MeshInfo
from esmf_regrid.tests import make_grid_args
from esmf_regrid.tests.unit.experimental.unstructured_regrid.test_MeshInfo import (
    _make_small_mesh_args,
)


def test_expanded_lons_with_mesh():
    """
    Basic test for regridding with :meth:`~esmf_regrid.esmf_regridder.RefinedGridInfo.make_esmf_field`.

    Mirrors the tests in :func:`~esmf_regrid.tests.unit.experimental.unstructured_regrid.test_MeshInfo.test_regrid_with_mesh`
    but with slightly different expected values due to increased accuracy.
    """
    mesh_args = _make_small_mesh_args()
    mesh = MeshInfo(*mesh_args)

    grid_args = make_grid_args(2, 3)
    grid = RefinedGridInfo(*grid_args[2:4], resolution=4)

    mesh_to_grid_regridder = Regridder(mesh, grid)
    mesh_input = np.array([3, 2])
    grid_output = mesh_to_grid_regridder.regrid(mesh_input)
    expected_grid_output = np.array(
        [
            [2.671534474734418, 3.0],
            [2.088765949748455, 2.922517356506756],
            [2.0, 2.340882413622917],
        ]
    )
    assert ma.allclose(expected_grid_output, grid_output)

    grid_to_mesh_regridder = Regridder(grid, mesh)
    grid_input = np.array([[0, 0], [1, 0], [2, 1]])
    mesh_output = grid_to_mesh_regridder.regrid(grid_input)
    expected_mesh_output = np.array([0.14117205318254747, 1.1976140197893996])
    assert ma.allclose(expected_mesh_output, mesh_output)

    def _give_extra_dims(array):
        result = np.stack([array, array + 1])
        result = np.stack([result, result + 10, result + 100])
        return result

    extra_dim_mesh_input = _give_extra_dims(mesh_input)
    extra_dim_grid_output = mesh_to_grid_regridder.regrid(extra_dim_mesh_input)
    extra_dim_expected_grid_output = _give_extra_dims(expected_grid_output)
    assert ma.allclose(extra_dim_expected_grid_output, extra_dim_grid_output)

    extra_dim_grid_input = _give_extra_dims(grid_input)
    extra_dim_mesh_output = grid_to_mesh_regridder.regrid(extra_dim_grid_input)
    extra_dim_expected_mesh_output = _give_extra_dims(expected_mesh_output)
    assert ma.allclose(extra_dim_expected_mesh_output, extra_dim_mesh_output)


def test_expanded_lats_with_mesh():
    """Basic test for regridding with :meth:`~esmf_regrid.esmf_regridder.RefinedGridInfo.make_esmf_field`."""
    mesh_args = _make_small_mesh_args()
    mesh = MeshInfo(*mesh_args)

    grid = RefinedGridInfo(np.array([0, 5, 10]), np.array([-90, 90]), resolution=4)

    mesh_to_grid_regridder = Regridder(mesh, grid)
    mesh_input = np.array([3, 2])
    grid_output = mesh_to_grid_regridder.regrid(mesh_input)
    expected_grid_output = np.array(
        [
            [2.2024695514629724, 2.4336888097502642],
        ]
    )
    assert ma.allclose(expected_grid_output, grid_output)

    grid_to_mesh_regridder = Regridder(grid, mesh)
    grid_input = np.array([[1, 2]])
    mesh_output = grid_to_mesh_regridder.regrid(grid_input)
    expected_mesh_output = np.array([1.7480791292591336, 1.496070008348207])
    assert ma.allclose(expected_mesh_output, mesh_output)

    def _give_extra_dims(array):
        result = np.stack([array, array + 1])
        result = np.stack([result, result + 10, result + 100])
        return result

    extra_dim_mesh_input = _give_extra_dims(mesh_input)
    extra_dim_grid_output = mesh_to_grid_regridder.regrid(extra_dim_mesh_input)
    extra_dim_expected_grid_output = _give_extra_dims(expected_grid_output)
    assert ma.allclose(extra_dim_expected_grid_output, extra_dim_grid_output)

    extra_dim_grid_input = _give_extra_dims(grid_input)
    extra_dim_mesh_output = grid_to_mesh_regridder.regrid(extra_dim_grid_input)
    extra_dim_expected_mesh_output = _give_extra_dims(expected_mesh_output)
    assert ma.allclose(extra_dim_expected_mesh_output, extra_dim_mesh_output)



"""Unit tests for :func:`esmf_regrid.experimental.unstructured_scheme.regrid_unstructured_to_unstructured`."""

from iris.coords import AuxCoord, DimCoord
from iris.cube import Cube
import numpy as np
import pytest

from esmf_regrid.experimental.unstructured_scheme import (
    regrid_unstructured_to_unstructured,
)
from esmf_regrid.tests.unit.schemes.test__mesh_to_MeshInfo import (
    _gridlike_mesh_cube,
)
from esmf_regrid.tests.unit.schemes.test__regrid_unstructured_to_rectilinear__prepare import (
    _flat_mesh_cube,
    _full_mesh,
)


def _add_metadata(cube):
    result = cube.copy()
    result.units = "K"
    result.attributes = {"a": 1}
    result.standard_name = "air_temperature"
    scalar_height = AuxCoord([5], units="m", standard_name="height")
    scalar_time = DimCoord([10], units="s", standard_name="time")
    result.add_aux_coord(scalar_height)
    result.add_aux_coord(scalar_time)
    return result


def test_flat_cubes():
    """
    Basic test for :func:`esmf_regrid.experimental.unstructured_scheme.regrid_unstructured_to_unstructured`.

    Tests with flat cubes as input (a 1D mesh cube and a 2D grid cube).
    """
    src = _flat_mesh_cube()

    n_lons = 6
    n_lats = 5
    tgt = _flat_mesh_cube()
    # Ensure data in the target grid is different to the expected data.
    # i.e. target grid data is all zero, expected data is all one
    tgt.data[:] = 0

    src = _add_metadata(src)
    src.data[:] = 1  # Ensure all data in the source is one.
    result = regrid_unstructured_to_unstructured(src, tgt)

    expected_data = np.ones([n_lats, n_lons])
    expected_cube = _add_metadata(tgt)

    # Lenient check for data.
    assert np.allclose(expected_data, result.data)

    # Check metadata and scalar coords.
    expected_cube.data = result.data
    assert expected_cube == result


@pytest.mark.parametrize("method", ("bilinear", "nearest"))
def test_node_friendly_methods(method):
    """
    Basic test for :func:`esmf_regrid.experimental.unstructured_scheme.regrid_unstructured_to_unstructured`.

    Tests with the bilinear and nearest method.
    """
    n_lons = 6
    n_lats = 5
    src = _gridlike_mesh_cube(n_lons, n_lats, location="node")
    tgt = _gridlike_mesh_cube(n_lons, n_lats, location="node")
    # Ensure data in the target mesh is different to the expected data.
    # i.e. target mesh data is all zero, expected data is all one
    tgt.data[:] = 0

    src = _add_metadata(src)
    src.data[:] = 1  # Ensure all data in the source is one.
    result = regrid_unstructured_to_unstructured(src, tgt, method=method)

    expected_data = np.ones_like(tgt.data)
    expected_cube = _add_metadata(tgt)

    # Lenient check for data.
    assert np.allclose(expected_data, result.data)

    # Check metadata and scalar coords.
    expected_cube.data = result.data
    assert expected_cube == result


def test_invalid_args():
    """
    Test for :func:`esmf_regrid.experimental.unstructured_scheme.regrid_unstructured_to_unstructured`.

    Tests that an appropriate error is raised when arguments are invalid.
    """
    n_lons = 6
    n_lats = 5
    node_src = _gridlike_mesh_cube(n_lons, n_lats, location="node")
    edge_src = _gridlike_mesh_cube(n_lons, n_lats, location="edge")
    face_src = _gridlike_mesh_cube(n_lons, n_lats, location="face")
    tgt = _gridlike_mesh_cube(n_lons, n_lats)

    with pytest.raises(ValueError):
        _ = regrid_unstructured_to_unstructured(face_src, tgt, method="other")
    with pytest.raises(ValueError) as excinfo:
        _ = regrid_unstructured_to_unstructured(node_src, tgt, method="conservative")
    expected_message = (
        "conservative regridding requires a source cube located on "
        "the face of a cube, target cube had the node location."
    )
    assert expected_message in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        _ = regrid_unstructured_to_unstructured(edge_src, tgt, method="bilinear")
    expected_message = (
        "bilinear regridding requires a source cube with a node "
        "or face location, target cube had the edge location."
    )
    assert expected_message in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        _ = regrid_unstructured_to_unstructured(edge_src, tgt, method="nearest")
    expected_message = (
        "nearest regridding requires a source cube with a node "
        "or face location, target cube had the edge location."
    )
    assert expected_message in str(excinfo.value)


def test_multidim_cubes():
    """
    Test for :func:`esmf_regrid.experimental.unstructured_scheme.regrid_unstructured_to_unstructured`.

    Tests with multidimensional cubes. The source cube contains
    coordinates on the dimensions before and after the mesh dimension.
    """

    mesh = _full_mesh()
    mesh_length = mesh.connectivity(contains_face=True).shape[0]

    h = 2
    t = 3
    height = DimCoord(np.arange(h), standard_name="height")
    time = DimCoord(np.arange(t), standard_name="time")

    src_data = np.empty([t, mesh_length, h])
    src_data[:] = np.arange(t * h).reshape([t, h])[:, np.newaxis, :]
    cube = Cube(src_data)
    mesh_coord_x, mesh_coord_y = mesh.to_MeshCoords("face")
    cube.add_aux_coord(mesh_coord_x, 1)
    cube.add_aux_coord(mesh_coord_y, 1)
    cube.add_dim_coord(time, 0)
    cube.add_dim_coord(height, 2)

    n_lons = 6
    n_lats = 5
    tgt = _gridlike_mesh_cube(n_lons, n_lats)

    result = regrid_unstructured_to_unstructured(cube, tgt)

    # Lenient check for data.
    expected_data = np.empty([t, n_lats * n_lons, h])
    expected_data[:] = np.arange(t * h).reshape(t, h)[:, np.newaxis, :]
    assert np.allclose(expected_data, result.data)

    expected_cube = Cube(expected_data)
    expected_cube.add_dim_coord(time, 0)
    expected_cube.add_aux_coord(tgt.coord("latitude"), 1)
    expected_cube.add_aux_coord(tgt.coord("longitude"), 1)
    expected_cube.add_dim_coord(height, 2)

    # Check metadata and scalar coords.
    result.data = expected_data
    assert expected_cube == result
